import sys
import os
import shutil
import itertools as itr
import segyio
import segyio.su as su

def product(f):
    return itr.product(range(len(f.ilines)), range(len(f.xlines)))

def pathjoin(prefix, path):
    dir, base = os.path.split(path)
    return os.path.join(dir, '-'.join((prefix, base)))

# this program copies the source-file and creates eight copies, each with a
# modified set of CDP-X and CDP-Y coordinates, rotating the field around the
# north (increasing CDP-Y) axis.
def main():
    if len(sys.argv) < 2:
        sys.exit("Usage: {} [source-file] [destination-file]".format(sys.argv[0]))

    srcfile = sys.argv[1]
    dstfile = sys.argv[2] if len(sys.argv) > 2 else srcfile

    for pre in ['normal', 'acute', 'right', 'obtuse', 'straight', 'reflex', 'left', 'inv-acute']:
        fname = pathjoin(pre, dstfile)
        shutil.copyfile(srcfile, fname)

        with segyio.open(srcfile) as src, segyio.open(fname, 'r+') as dst:
            for i in range(1 + src.ext_headers):
                dst.text[i] = src.text[i]

            dst.bin = src.bin
            dst.trace = src.trace
            dst.header = src.header

    with segyio.open(pathjoin('normal', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = x
            trh[su.cdpy] = y
            trh[su.scalco] = 10

    with segyio.open(pathjoin('acute', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[y + x * len(src.ilines)]
            trh[su.cdpx] = x + y
            trh[su.cdpy] = (100 - x) + y
            trh[su.scalco] = -10

    with segyio.open(pathjoin('right', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = y
            trh[su.cdpy] = 100 - x
            trh[su.scalco] = 1

    with segyio.open(pathjoin('obtuse', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = (100 - x) + y
            trh[su.cdpy] = (100 - x) - y
            trh[su.scalco] = 2

    with segyio.open(pathjoin('straight', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = 100 - x
            trh[su.cdpy] = 100 - y
            trh[su.scalco] = -7

    with segyio.open(pathjoin('reflex', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = 100 - (x + y)
            trh[su.cdpy] = 100 + (x - y)
            trh[su.scalco] = 7

    with segyio.open(pathjoin('left', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = 100 - y
            trh[su.cdpy] = x
            trh[su.scalco] = 21

    with segyio.open(pathjoin('inv-acute', dstfile), 'r+') as dst:
        for i, (x, y) in enumerate(product(src)):
            trh = dst.header[i]
            trh[su.cdpx] = 100 + x - y
            trh[su.cdpy] = x + y
            trh[su.scalco] = 100

if __name__ == '__main__':
    main()



from __future__ import absolute_import

import os
import unittest

from .test_context import TestContext


class TestContextTest(unittest.TestCase):
    def test_test_context(self):
        original_cwd = os.getcwd()
        with TestContext() as context:
            cwd = os.getcwd()
            self.assertNotEqual(original_cwd, cwd)
            self.assertEqual(original_cwd, context.cwd)
            self.assertEqual(cwd, context.temp_path)
            self.assertTrue(cwd.endswith("_unnamed"))
            context_path = context.temp_path

        self.assertFalse(os.path.exists(context_path))
        self.assertEqual(original_cwd, os.getcwd())

    def test_named_test_context(self):
        with TestContext(name="named") as context:
            self.assertTrue(context.temp_path.endswith("_named"))

    def test_test_context_copy(self):
        with TestContext() as context:
            small = "test-data/small.sgy"
            context.copy_file(small)
            context.copy_file(small, "test")

            self.assertTrue(os.path.exists("small.sgy"))
            self.assertTrue(os.path.exists("test/small.sgy"))

if __name__ == '__main__':
    unittest.main()



import sys
import segyio

# this program creates a new subcube, taking the first 5 lines in both
# directions, and reduces the trace size to 20 samples
def main():
    if len(sys.argv) < 3:
        sys.exit("Usage: {} [source-file] [destination-file]".format(sys.argv[0]))

    sourcefile = sys.argv[1]
    destfile = sys.argv[2]

    with segyio.open(sourcefile) as src:
        spec = segyio.spec()
        spec.sorting = int(src.sorting)
        spec.format  = int(src.format)
        spec.samples = range(50)
        spec.ilines = src.ilines[:5]
        spec.xlines = src.xlines[:5]

        with segyio.create(destfile, spec) as dst:
            # Copy all textual headers, including possible extended
            for i in range(1 + src.ext_headers):
                dst.text[i] = src.text[i]

            # copy the binary header, then insert the modifications needed for
            # the new shape
            dst.bin = src.bin
            dst.bin = { segyio.BinField.Samples: 50,
                        segyio.BinField.Traces: 5 * 5
                      }

            # Copy all headers in the new inlines. Since we know we're copying
            # the five first we don't have to take special care to update
            # headers
            dst.header.iline = src.header.iline
            # the copy traces (in line mode)
            dst.iline = src.iline

if __name__ == '__main__':
    main()



import unittest
import subprocess

class appTest(unittest.TestCase):
    def setUp(self):
        self.file_path = "../python/test-data/small.sgy"
        self.cmd_base = ".././applications/segyio-catr"
        self.ref_list = [
                    b'tracl\t0',
                    b'tracr\t0',
                    b'fldr\t0',
                    b'tracf\t0',
                    b'ep\t0',
                    b'cdp\t0',
                    b'cdpt\t0',
                    b'trid\t0',
                    b'nvs\t0',
                    b'nhs\t0',
                    b'duse\t0',
                    b'offset\t1',
                    b'gelev\t0',
                    b'selev\t0',
                    b'sdepth\t0',
                    b'gdel\t0',
                    b'sdel\t0',
                    b'swdep\t0',
                    b'gwdep\t0',
                    b'scalel\t0',
                    b'scalco\t0',
                    b'sx\t0',
                    b'sy\t0',
                    b'gx\t0',
                    b'gy\t0',
                    b'counit\t0',
                    b'wevel\t0',
                    b'swevel\t0',
                    b'sut\t0',
                    b'gut\t0',
                    b'sstat\t0',
                    b'gstat\t0',
                    b'tstat\t0',
                    b'laga\t0',
                    b'lagb\t0',
                    b'delrt\t0',
                    b'muts\t0',
                    b'mute\t0',
                    b'ns\t0',
                    b'dt\t0',
                    b'gain\t0',
                    b'igc\t0',
                    b'igi\t0',
                    b'corr\t0',
                    b'sfs\t0',
                    b'sfe\t0',
                    b'slen\t0',
                    b'styp\t0',
                    b'stat\t0',
                    b'stae\t0',
                    b'tatyp\t0',
                    b'afilf\t0',
                    b'afils\t0',
                    b'nofilf\t0',
                    b'nofils\t0',
                    b'lcf\t0',
                    b'hcf\t0',
                    b'lcs\t0',
                    b'hcs\t0',
                    b'year\t0',
                    b'day\t0',
                    b'hour\t0',
                    b'minute\t0',
                    b'sec\t0',
                    b'timbas\t0',
                    b'trwf\t0',
                    b'grnors\t0',
                    b'grnofr\t0',
                    b'grnlof\t0',
                    b'gaps\t0',
                    b'otrav\t0',
                    b'cdpx\t0',
                    b'cdpy\t0',
                    b'iline\t2',
                    b'xline\t20',
                    b'sp\t0',
                    b'scalsp\t0',
                    b'trunit\t0',
                    b'tdcm\t0',
                    b'tdcp\t0',
                    b'tdunit\t0',
                    b'triden\t0',
                    b'sctrh\t0',
                    b'stype\t0',
                    b'sedm\t0',
                    b'sede\t0',
                    b'smm\t0',
                    b'sme\t0',
                    b'smunit\t0',
                    b'uint1\t0',
                    b'uint2\t0']

    def test_segycatr_output_t(self):
        cmd = [self.cmd_base, self.file_path, "-t 6"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i])

    def test_segycatr_output_r1(self):
        cmd = [self.cmd_base, self.file_path, "-r 6"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i])

    def test_segycatr_output_r2(self):
        cmd = [self.cmd_base, self.file_path, "-r 1 6"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i+455])

    def test_segycatr_output_r3(self):
        cmd = [self.cmd_base, self.file_path, "-r 5 7 1"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i+91])

    def test_segycatr_output_r_multiple(self):
        cmd = [self.cmd_base, self.file_path, "-r 1 2", "-r 6 7"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i+182])
            
    def test_segycatr_output_r_t(self):
        cmd = [self.cmd_base, self.file_path, "-r 4 9", "-t 12"]
        res = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
        res_list = res.split(b"\n")
        for i in range(len(self.ref_list)):
            self.assertEquals(self.ref_list[i], res_list[i+182])

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import division
import unittest

import math
from segyio import BinField
from segyio import TraceField
import numpy as np
from .test_context import TestContext
import segyio


class ToolsTest(unittest.TestCase):
    def setUp(self):
        self.filename = "test-data/small.sgy"
        self.prestack = "test-data/small-ps.sgy"

    def test_dt_fallback(self):
        with TestContext("dt_fallback") as context:
            context.copy_file(self.filename)
            with segyio.open("small.sgy", "r+") as f:
                # Both zero
                f.bin[BinField.Interval] = 0
                f.header[0][TraceField.TRACE_SAMPLE_INTERVAL] = 0
                f.flush()
                fallback_dt = 4
                np.testing.assert_almost_equal(segyio.dt(f, fallback_dt), fallback_dt)

                # dt in bin header different from first trace
                f.bin[BinField.Interval] = 6000
                f.header[0][TraceField.TRACE_SAMPLE_INTERVAL] = 1000
                f.flush()
                fallback_dt = 4
                np.testing.assert_almost_equal(segyio.dt(f, fallback_dt), fallback_dt)

    def test_dt_no_fallback(self):
        with TestContext("dt_no_fallback") as context:
            context.copy_file(self.filename)
            dt_us = 6000
            with segyio.open("small.sgy", "r+") as f:
                f.bin[BinField.Interval] = dt_us
                f.header[0][TraceField.TRACE_SAMPLE_INTERVAL] = dt_us
                f.flush()
                np.testing.assert_almost_equal(segyio.dt(f), dt_us)

    def test_sample_indexes(self):
        with segyio.open(self.filename, "r") as f:
            indexes = segyio.sample_indexes(f)
            step = 4000.0
            self.assertListEqual(indexes, [t * step for t in range(len(f.samples))])

            indexes = segyio.sample_indexes(f, t0=1.5)
            self.assertListEqual(indexes, [1.5 + t * step for t in range(len(f.samples))])

            indexes = segyio.sample_indexes(f, t0=1.5, dt_override=3.21)
            self.assertListEqual(indexes, [1.5 + t * 3.21 for t in range(len(f.samples))])

    def test_empty_text_header_creation(self):
        text_header = segyio.create_text_header({})

        for line_no in range(0, 40):
            line = text_header[line_no * 80: (line_no + 1) * 80]
            self.assertEqual(line, "C{0:>2} {1:76}".format(line_no + 1, ""))

    def test_wrap(self):
        with segyio.open(self.filename, "r") as f:
            segyio.tools.wrap(f.text[0])
            segyio.tools.wrap(f.text[0], 90)

    def test_values_text_header_creation(self):
        lines = {i + 1: chr(64 + i) * 76 for i in range(40)}
        text_header = segyio.create_text_header(lines)

        for line_no in range(0, 40):
            line = text_header[line_no * 80: (line_no + 1) * 80]
            self.assertEqual(line, "C{0:>2} {1:76}".format(line_no + 1, chr(64 + line_no) * 76))

    def test_native(self):
        with open(self.filename, 'rb') as f, segyio.open(self.filename) as sgy:
            f.read(3600+240)
            filetr = f.read(4 * len(sgy.samples))
            segytr = sgy.trace[0]

            filetr = np.frombuffer(filetr, dtype = np.single)
            self.assertFalse(np.array_equal(segytr, filetr))
            self.assertTrue(np.array_equal(segytr, segyio.tools.native(filetr)))

    def test_cube_filename(self):
        with segyio.open(self.filename) as f:
            c1 = segyio.tools.cube(f)
            c2 = segyio.tools.cube(self.filename)
            self.assertTrue(np.all(c1 == c2))

    def test_cube_identity(self):
        with segyio.open(self.filename) as f:
            x = segyio.tools.collect(f.trace[:])
            x = x.reshape((len(f.ilines), len(f.xlines), len(f.samples)))
            self.assertTrue(np.all(x == segyio.tools.cube(f)))

    def test_cube_identity_prestack(self):
        with segyio.open(self.prestack) as f:
            dims = (len(f.ilines), len(f.xlines), len(f.offsets), len(f.samples))
            x = segyio.tools.collect(f.trace[:]).reshape(dims)
            self.assertTrue(np.all(x == segyio.tools.cube(f)))

    def test_unstructured_rotation(self):
        with self.assertRaises(ValueError):
            with segyio.open(self.filename, ignore_geometry = True) as f:
                segyio.tools.rotation(f)

    def test_rotation(self):
        names = ['normal', 'acute', 'right', 'obtuse',
                 'straight', 'reflex', 'left', 'inv-acute']
        angles = [0.000, 0.785, 1.571, 2.356,
                  3.142, 3.927, 4.712, 5.498]
        rights = [1.571, 2.356, 3.142, 3.927,
                  4.712, 5.498, 0.000, 0.785 ]

        def rotation(x, **kwargs): return segyio.tools.rotation(x, **kwargs)[0]
        close = self.assertAlmostEqual

        for name, angle, right in zip(names, angles, rights):
            src = self.filename.replace('/', '/' + name + '-')

            with segyio.open(src) as f:
                close(angle, rotation(f), places = 3)
                close(angle, rotation(f, line = 'fast'),  places = 3)
                close(angle, rotation(f, line = 'iline'), places = 3)
                close(right, rotation(f, line = 'slow'),  places = 3)
                close(right, rotation(f, line = 'xline'), places = 3)

if __name__ == '__main__':
    unittest.main()



from unittest.case import TestCase

from segyio import Enum


class TestEnum(Enum):
    ZERO = 0
    ONE = 1
    TWO = 2


class EnumTest(TestCase):

    def test_enum(self):
        self.assertEqual(TestEnum.ONE, 1)

        one = TestEnum(TestEnum.ONE)
        self.assertIsInstance(one, TestEnum)
        self.assertEqual(str(one), "ONE")
        self.assertEqual(int(one), 1)
        self.assertEqual(one, 1)

        self.assertListEqual(TestEnum.enums(), TestEnum.enums())
        self.assertListEqual(TestEnum.enums(), [0, 1, 2])



import numpy as np
import segyio
from segyio._raw_trace import RawTrace
import itertools

try:
    from itertools import izip as zip
    from itertools import imap as map
except ImportError:  # will be 3.x series
    pass


class Trace:

    def __init__(self, file):
        self._file = file
        """:type: segyio.file"""

    def __getitem__(self, index, buf=None):
        if isinstance(index, tuple):
            return self.__getitem__(index[0], index[1])

        buf = self._trace_buffer(buf)

        if isinstance(index, slice):
            # always read the trace into a second buffer. This is to provide
            # exception safety: if an exception is raised and at least one
            # array has already been yielded to the caller, failing to read the
            # next trace won't make the already-returned array garbage
            def gen():
                buf1 = buf
                buf2 = self._trace_buffer(None)
                for i in range(*index.indices(len(self))):
                    buf1, buf2 = self._readtr(i, 1, 1, buf1, buf2)
                    yield buf1

            return gen()

        if not 0 <= abs(index) < len(self):
            raise IndexError("Trace %d not in range (-%d,%d)", (index, len(self), len(self)))

        # map negative a negative to the corresponding positive value
        start = (index + len(self)) % len(self)
        return self._readtr(start, 1, 1, buf)[0]

    def __setitem__(self, index, val):
        if isinstance(index, slice):
            for i, x in zip(range(*index.indices(len(self))), val):
                self.write_trace(i, x, self._file)
            return

        if int(index) != index:
            raise TypeError("Trace index must be integer type")

        if not 0 <= abs(index) < len(self):
            raise IndexError("Trace %d not in range (-%d,%d)",
                             (index, len(self), len(self)))

        self.write_trace(index, val, self._file)

    def __len__(self):
        return self._file.tracecount

    def __iter__(self):
        return self[:]

    def __repr__(self):
        return "Trace(traces = {}, samples = {})".format(len(self), self._file.samples)

    def _trace_buffer(self, buf=None):
        shape = self._file.samples.shape

        if buf is None:
            buf = np.empty(shape=shape, dtype=np.single)
        elif not isinstance(buf, np.ndarray):
            raise TypeError("Buffer must be None or numpy.ndarray")
        elif buf.dtype != np.single:
            buf = np.empty(shape=shape, dtype=np.single)

        return buf

    def _readtr(self, start, step, length, buf, buf1 = None):
        if buf1 is None:
            buf1 = buf

        trace0 = self._file._tr0
        bsz = self._file._bsz
        fmt = self._file._fmt
        smp = len(self._file.samples)

        buf1 = segyio._segyio.read_trace(self._file.xfd, buf1,
                                         start, step, length,
                                         fmt, smp,
                                         trace0, bsz)

        return buf1, buf

    @classmethod
    def write_trace(cls, traceno, buf, segy):
        """
        :type traceno: int
        :type buf: ?
        :type segy: segyio.SegyFile
        """
        if isinstance(buf, np.ndarray) and buf.dtype != np.single:
            raise TypeError("Numpy array must be of type single")

        segyio._segyio.write_trace(segy.xfd, traceno,
                                   buf,
                                   segy._tr0, segy._bsz,
                                   segy._fmt, len(segy.samples))

    @property
    def raw(self):
        """ :rtype: segyio.RawTrace """
        return RawTrace(self)



from segyio import Enum
from segyio import TraceField as tf
from segyio import BinField as bf

class su(Enum):
    """ Since v1.1
        Seismic Unix style aliases for binary and trace header fields.

    About names:
        Seismic Unix does not have names for all possible fields, as it came
        around during an early revision of SEG-Y, and names for these fields
        are created by segyio in their absence. If Seismic Unix starts
        providing names for these fields, they will be added to these alies,
        and in conflicts take presedence after some time. If there are no
        conflicts, segyio names are considered for deprecation on a
        case-by-case basis, but will most likely be supported along with the
        Seismic Unix name.
    """
# trace
    tracl  = tf.TRACE_SEQUENCE_LINE
    tracr  = tf.TRACE_SEQUENCE_FILE
    fldr   = tf.FieldRecord
    tracf  = tf.TraceNumber
    ep     = tf.EnergySourcePoint
    cdp    = tf.CDP
    cdpt   = tf.CDP_TRACE
    trid   = tf.TraceIdentificationCode
    nvs    = tf.NSummedTraces
    nhs    = tf.NStackedTraces
    duse   = tf.DataUse
    offset = tf.offset
    gelev  = tf.ReceiverGroupElevation
    selev  = tf.SourceSurfaceElevation
    sdepth = tf.SourceDepth
    gdel   = tf.ReceiverDatumElevation
    sdel   = tf.SourceDatumElevation
    swdep  = tf.SourceWaterDepth
    gwdep  = tf.GroupWaterDepth
    scalel = tf.ElevationScalar
    scalco = tf.SourceGroupScalar
    sx     = tf.SourceX
    sy     = tf.SourceY
    gx     = tf.GroupX
    gy     = tf.GroupY
    counit = tf.CoordinateUnits
    wevel  = tf.WeatheringVelocity
    swevel = tf.SubWeatheringVelocity
    sut    = tf.SourceUpholeTime
    gut    = tf.GroupUpholeTime
    sstat  = tf.SourceStaticCorrection
    gstat  = tf.GroupStaticCorrection
    tstat  = tf.TotalStaticApplied
    laga   = tf.LagTimeA
    lagb   = tf.LagTimeB
    delrt  = tf.DelayRecordingTime
    muts   = tf.MuteTimeStart
    mute   = tf.MuteTimeEND
    ns     = tf.TRACE_SAMPLE_COUNT
    dt     = tf.TRACE_SAMPLE_INTERVAL
    gain   = tf.GainType
    igc    = tf.InstrumentGainConstant
    igi    = tf.InstrumentInitialGain
    corr   = tf.Correlated
    sfs    = tf.SweepFrequencyStart
    sfe    = tf.SweepFrequencyEnd
    slen   = tf.SweepLength
    styp   = tf.SweepType
    stat   = tf.SweepTraceTaperLengthStart
    stae   = tf.SweepTraceTaperLengthEnd
    tatyp  = tf.TaperType
    afilf  = tf.AliasFilterFrequency
    afils  = tf.AliasFilterSlope
    nofilf = tf.NotchFilterFrequency
    nofils = tf.NotchFilterSlope
    lcf    = tf.LowCutFrequency
    hcf    = tf.HighCutFrequency
    lcs    = tf.LowCutSlope
    hcs    = tf.HighCutSlope
    year   = tf.YearDataRecorded
    day    = tf.DayOfYear
    hour   = tf.HourOfDay
    minute = tf.MinuteOfHour
    sec    = tf.SecondOfMinute
    timbas = tf.TimeBaseCode
    trwf   = tf.TraceWeightingFactor
    grnors = tf.GeophoneGroupNumberRoll1
    grnofr = tf.GeophoneGroupNumberFirstTraceOrigField
    grnlof = tf.GeophoneGroupNumberLastTraceOrigField
    gaps   = tf.GapSize
    otrav  = tf.OverTravel
    cdpx   = tf.CDP_X
    cdpy   = tf.CDP_Y
    iline  = tf.INLINE_3D
    xline  = tf.CROSSLINE_3D
    sp     = tf.ShotPoint
    scalsp = tf.ShotPointScalar
    trunit = tf.TraceValueMeasurementUnit
    tdcm   = tf.TransductionConstantMantissa
    tdcp   = tf.TransductionConstantPower
    tdunit = tf.TransductionUnit
    triden = tf.TraceIdentifier
    sctrh  = tf.ScalarTraceHeader
    stype  = tf.SourceType
    sedm   = tf.SourceEnergyDirectionMantissa
    sede   = tf.SourceEnergyDirectionExponent
    smm    = tf.SourceMeasurementMantissa
    sme    = tf.SourceMeasurementExponent
    smunit = tf.SourceMeasurementUnit
    uint1  = tf.UnassignedInt1
    uint2  = tf.UnassignedInt2

# binary
    jobid  = bf.JobID
    lino   = bf.LineNumber
    reno   = bf.ReelNumber
    ntrpr  = bf.Traces
    nart   = bf.AuxTraces
    hdt    = bf.Interval
    dto    = bf.IntervalOriginal
    hns    = bf.Samples
    nso    = bf.SamplesOriginal
    format = bf.Format
    fold   = bf.EnsembleFold
    tsort  = bf.SortingCode
    vscode = bf.VerticalSum
    hsfs   = bf.SweepFrequencyStart
    hsfe   = bf.SweepFrequencyEnd
    hslen  = bf.SweepLength
    hstyp  = bf.Sweep
    schn   = bf.SweepChannel
    hstas  = bf.SweepTaperStart
    hstae  = bf.SweepTaperEnd
    htatyp = bf.Taper
    hcorr  = bf.CorrelatedTraces
    bgrcv  = bf.BinaryGainRecovery
    rcvm   = bf.AmplitudeRecovery
    mfeet  = bf.MeasurementSystem
    polyt  = bf.ImpulseSignalPolarity
    vpol   = bf.VibratoryPolarity
    unas1  = bf.Unassigned1
    rev    = bf.SEGYRevision
    trflag = bf.TraceFlag
    exth   = bf.ExtendedHeaders
    unas2  = bf.Unassigned2



import segyio
from segyio import BinField
from segyio import TraceField

class Field(object):
    _bin_keys = [x for x in BinField.enums()
                 if  x != BinField.Unassigned1
                 and x != BinField.Unassigned2]

    _tr_keys = [x for x in TraceField.enums()
                if  x != TraceField.UnassignedInt1
                and x != TraceField.UnassignedInt2]


    def __init__(self, buf, write, field_type, traceno=None, keys = _tr_keys):
        self.buf = buf
        self.traceno = traceno
        self._field_type = field_type
        self._keys = keys
        self._write = write

    def _get_field(self, *args):
        return segyio._segyio.get_field(self.buf, *args)

    def _set_field(self, *args):
        return segyio._segyio.set_field(self.buf, *args)

    def __getitem__(self, field):
        # add some structure so we can always iterate over fields
        if isinstance(field, int) or isinstance(field, self._field_type):
            field = [field]

        d = {self._field_type(f): self._get_field(f) for f in field}

        # unpack the dictionary. if header[field] is requested, a
        # plain, unstructed output is expected, but header[f1,f2,f3]
        # yields a dict
        if len(d) == 1:
            return d.popitem()[1]

        return d

    def keys(self):
        return list(self._keys)

    def values(self):
        return map(self._get_field, self.keys())

    def items(self):
        return zip(self.keys(), self.values())

    def __contains__(self, key):
        return key in self._keys

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return iter(self._keys)

    def __setitem__(self, field, val):
        self._set_field(field, val)
        self._write(self.buf, self.traceno)

    def update(self, value):
        if type(self) is type(value):
            buf = value.buf
        else:
            buf = self.buf

            # iter() on a dict only gives values, need key-value pairs
            try: value = value.items()
            except AttributeError: pass

            for k, v in value:
                self._set_field(int(k), v)

        self._write(buf, self.traceno)

    @classmethod
    def binary(cls, segy):
        try:
            buf = segyio._segyio.read_binaryheader(segy.xfd)
        except IOError:
            # the file was probably newly created and the binary header hasn't
            # been written yet.  if this is the case we want to try and write
            # it. if the file was broken, permissions were wrong etc writing
            # will fail too
            buf = segyio._segyio.empty_binaryheader()

        def wr(buf, *_):
            segyio._segyio.write_binaryheader(segy.xfd, buf)

        return Field(buf, write=wr, field_type=BinField, keys = Field._bin_keys)

    @classmethod
    def trace(cls, buf, traceno, segy):
        if traceno < 0:
            traceno = segy.tracecount + traceno

        if traceno >= segy.tracecount or traceno < 0:
            raise IndexError("Header out of range: 0 <= %d < %d" % (traceno, segy.tracecount))

        if buf is None:
            buf = segyio._segyio.empty_traceheader()

        try:
            segyio._segyio.read_traceheader(segy.xfd, traceno, buf, segy._tr0, segy._bsz)
        except IOError:
            # the file was probably newly created and the trace header hasn't
            # been written yet.  if this is the case we want to try and write
            # it. if the file was broken, permissions were wrong etc writing
            # will fail too
            pass

        def wr(buf, traceno):
            segyio._segyio.write_traceheader(segy.xfd, traceno, buf, segy._tr0, segy._bsz)

        return Field(buf, traceno=traceno,
                          write=wr,
                          field_type=TraceField,
                          keys=Field._tr_keys)

    def __repr__(self):
        return self[self.keys()].__repr__()



import numpy as np
import segyio

try: xrange
except NameError: pass
else: range = xrange


class RawTrace(object):
    def __init__(self, trace):
        self.trace = trace

    def __getitem__(self, index):
        """ :rtype: numpy.ndarray """
        buf = None
        if isinstance(index, slice):
            f = self.trace._file
            start, stop, step = index.indices(f.tracecount)
            mstart, mstop = min(start, stop), max(start, stop)
            length = max(0, (mstop - mstart + (step - (1 if step > 0 else -1))))
            buf = np.zeros(shape = (length, len(f.samples)), dtype = np.single)
            l = len(range(start, stop, step))
            buf, _ = self.trace._readtr(start, step, l, buf)
            return buf

        if int(index) != index:
            raise TypeError("Trace index must be integer or slice.")

        buf = self.trace._trace_buffer(None)
        return self.trace._readtr(int(index), 1, 1, buf)[0]

    def __repr__(self):
        return self.trace.__repr__() + ".raw"



import sys
import numpy as np
import segyio


def main():
    if len(sys.argv) < 7:
        sys.exit("Usage: {} [file] [samples] [first iline] [last iline] [first xline] [last xline]".format(sys.argv[0]))

    spec = segyio.spec()
    filename = sys.argv[1]

    # to create a file from nothing, we need to tell segyio about the structure of
    # the file, i.e. its inline numbers, crossline numbers, etc. You can also add
    # more structural information, but offsets etc. have sensible defautls. This is
    # the absolute minimal specification for a N-by-M volume
    spec.sorting = 2
    spec.format = 1
    spec.samples = range(int(sys.argv[2]))
    spec.ilines = range(*map(int, sys.argv[3:5]))
    spec.xlines = range(*map(int, sys.argv[5:7]))

    with segyio.create(filename, spec) as f:
        # one inline consists of 50 traces
        # which in turn consists of 2000 samples
        start = 0.0
        step = 0.00001
        # fill a trace with predictable values: left-of-comma is the inline
        # number. Immediately right of comma is the crossline number
        # the rightmost digits is the index of the sample in that trace meaning
        # looking up an inline's i's jth crosslines' k should be roughly equal
        # to i.j0k
        trace = np.arange(start = start,
                          stop  = start + step * len(spec.samples),
                          step  = step,
                          dtype = np.single)

        # one inline is N traces concatenated. We fill in the xline number
        line = np.concatenate([trace + (xl / 100.0) for xl in spec.xlines])
        line = line.reshape( (len(spec.xlines), len(spec.samples)) )

        # write the line itself to the file
        # write the inline number in all this line's headers
        for ilno in spec.ilines:
            f.iline[ilno] = (line + ilno)
            f.header.iline[ilno] = { segyio.TraceField.INLINE_3D: ilno,
                                     segyio.TraceField.offset: 1
                                   }

        # then do the same for xlines
        for xlno in spec.xlines:
            f.header.xline[xlno] = { segyio.TraceField.CROSSLINE_3D: xlno }


if __name__ == '__main__':
    main()



import sys
from segyio import TraceField
import segyio


def list_byte_offset_names():
    print("Available offsets and their corresponding byte value:")
    for x in TraceField.enums():
        print("  %s: %d" % (str(x), x))


if __name__ == '__main__':
    if len(sys.argv) < 4:
        list_byte_offset_names()
        sys.exit("Usage: about.py [file] [inline] [crossline]")

    # we need a way to convert from run-time inline/crossline argument (as
    # text) to the internally used TraceField enum. Make a string -> TraceField
    # map and look up into that. this dictionary comprehension creates that
    fieldmap = {str(x).lower(): x for x in TraceField.enums()}

    filename = sys.argv[1]
    inline_name, crossline_name = sys.argv[2].lower(), sys.argv[3].lower()

    # exit if inline or crossline are unknown
    if inline_name not in fieldmap:
        list_byte_offset_names()
        sys.exit("Unknown inline field '%s'" % sys.argv[2])

    if crossline_name not in fieldmap:
        list_byte_offset_names()
        sys.exit("Unknown crossline field '%s'" % sys.argv[3])

    inline, crossline = fieldmap[inline_name], fieldmap[crossline_name]

    with segyio.open(filename, "r", inline, crossline) as f:
        print("About '%s':" % filename)
        print("Format type: %s" % f.format)
        print("Offset count: %d" % f.offsets)
        print("ilines: %s" % ", ".join(map(str, f.ilines)))
        print("xlines: %s" % ", ".join(map(str, f.xlines)))

    print("+------+")

    with segyio.open(filename, "r", crossline, inline) as f:
        # with swapped inline/crossline
        print("About '%s':" % filename)
        print("Format type: %s" % f.format)
        print("Offset count: %d" % f.offsets)
        print("ilines: %s" % ", ".join(map(str, f.ilines)))
        print("xlines: %s" % ", ".join(map(str, f.xlines)))



import unittest
import subprocess
import os

class TestSegyB(unittest.TestCase):

    actual_output = [
        b"jobid\t0",
        b"lino\t0",
        b"reno\t0",
        b"ntrpr\t25",
        b"nart\t0",
        b"hdt\t4000",
        b"dto\t0",
        b"hns\t50",
        b"nso\t0",
        b"format\t1",
        b"fold\t0",
        b"tsort\t0",
        b"vscode\t0",
        b"hsfs\t0",
        b"hsfe\t0",
        b"hslen\t0",
        b"hstyp\t0",
        b"schn\t0",
        b"hstas\t0",
        b"hstae\t0",
        b"htatyp\t0",
        b"hcorr\t0",
        b"bgrcv\t0",
        b"rcvm\t0",
        b"mfeet\t0",
        b"polyt\t0",
        b"vpol\t0",
        b"rev\t0",
        b"trflag\t0",
        b"exth\t0"
    ]
    work_dir = '../'
    cmd_base = './applications/segyio-catb'
    FNULL = open(os.devnull, 'w')

    def test_segy_in(self):
        filepath = self.work_dir + '/python/test-data/small.sgy'
        cmd = self.cmd_base + ' python/test-data/small.sgy'
        result = subprocess.check_output(cmd, shell=True, cwd=self.work_dir)
        result_list = result.split(b'\n')
        for i in range(len(self.actual_output)):
            self.assertEqual(result_list[i], self.actual_output[i])

if __name__ == '__main__':
    unittest.main()




import sys
import segyio
import numpy as np


def main():
    if len(sys.argv) < 2:
        sys.exit("Usage: {} [segyfile] ".format(sys.argv[0]))

    segyfile = sys.argv[1]

    min_value = sys.float_info.max
    max_value = sys.float_info.min

    with segyio.open(segyfile) as f:
        for trace in f.trace:

            local_min = np.nanmin(trace)
            local_max = np.nanmax(trace)

            if np.isfinite(local_min):
                min_value = min(local_min, min_value)

            if np.isfinite(local_max):
                max_value = max(local_max, max_value)

    print("min: {}".format(min_value))
    print("max: {}".format(max_value))

if __name__ == '__main__':
    main()



import os
import tempfile

import shutil

import errno


class TestContext(object):
    def __init__(self, name="unnamed", cleanup=True):
        super(TestContext, self).__init__()
        self._name = name
        self._cwd = os.getcwd()
        self._cleanup = cleanup

    def __enter__(self):
        temp_path = tempfile.mkdtemp("_%s" % self._name)
        os.chdir(temp_path)
        self._temp_path = os.getcwd()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._cwd)
        if self._cleanup:
            shutil.rmtree(self._temp_path)

    def copy_file(self, source, target="."):
        if not os.path.isabs(source):
            source = os.path.join(self._cwd, source)

        source_dir, source_filename = os.path.split(source)

        target = os.path.join(self._temp_path, target, source_filename)
        try:
            os.makedirs(os.path.dirname(target))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        shutil.copyfile(source, target)

    @property
    def temp_path(self):
        return self._temp_path

    @property
    def cwd(self):
        return self._cwd


import itertools
import numpy as np

import segyio.tools as tools
from segyio.tracesortingformat import TraceSortingFormat

class Gather:
    """ Gather mode. Internal.

    Provides the implementation for f.gather, reading n offsets from the
    intersection of two lines in a cube.
    """

    def __init__(self, trace, iline, xline, offsets, sort):
        # cache constructed modes for performance
        self.trace     = trace
        self.iline     = iline
        self.xline     = xline
        self.offsets   = offsets
        self.sort      = sort

    def _getindex(self, il, xl, offset, sorting):
        """ Get the trace index for an (il, xl, offset) tuple

        :rtype: int
        """
        if sorting == TraceSortingFormat.INLINE_SORTING:
            xlind = self.xline._index(xl, None)
            return self.iline._index(il, offset) + xlind
        else:
            ilind = self.iline._index(il, None)
            return self.xline._index(xl, offset) + ilind

    def __getitem__(self, index):
        """ :rtype: iterator[numpy.ndarray]|numpy.ndarray """
        if len(index) < 3:
            index = (index[0], index[1], slice(None))

        il, xl, off = index
        sort = self.sort

        def isslice(x): return isinstance(x, slice)

        # gather[int,int,int]
        if not any(map(isslice, [il, xl, off])):
            return self.trace[self._getindex(il, xl, off, sort)]

        offs = off if isslice(off) else slice(off, off+1, 1)

        xs = list(filter(self.offsets.__contains__,
                    range(*offs.indices(self.offsets[-1]+1))))

        empty = np.empty(0, dtype = np.single)
        # gather[int,int,:]
        if not any(map(isslice, [il, xl])):
            if len(xs) == 0: return empty
            return tools.collect(
                    self.trace[self._getindex(il, xl, x, sort)]
                    for x in xs)

        # gather[:,:,:], gather[int,:,:], gather[:,int,:]
        # gather[:,:,int] etc
        def gen():
            # precompute the xline number -> xline offset
            xlinds = { xlno: i for i, xlno in enumerate(self.xline.lines) }
            ofinds = { ofno: i for i, ofno in enumerate(self.offsets) }

            # doing range over gathers is VERY expensive, because every lookup
            # with a naive implementations would call _getindex to map lineno
            # -> trace index. However, ranges over gathers are done on a
            # by-line basis so lines can be buffered, and traces can be read
            # from the iline. This is the least efficient when there are very
            # few traces read per inline, but huge savings with larger subcubes
            last_il = self.iline.lines[-1] + 1
            last_xl = self.xline.lines[-1] + 1

            il_slice = il if isslice(il) else slice(il, il+1)
            xl_slice = xl if isslice(xl) else slice(xl, xl+1)

            il_range = range(*il_slice.indices(last_il))
            xl_range = range(*xl_slice.indices(last_xl))

            if not isslice(off):
                for iline in self.iline[il_slice, off]:
                    for xlno in xl_range:
                        yield iline[xlinds[xlno]]

                return

            if len(xs) == 0:
                for _, _ in itertools.product(ilrange, xlrange): yield empty
                return

            for ilno in il_range:
                iline = tools.collect(self.iline[ilno, off])
                for x in xl_range:
                    yield iline[:, xlinds[x]]

        return gen()



import sys
import segyio
import numpy as np

def main():
    if len( sys.argv ) < 2:
        sys.exit("Usage: write.py [file]")

    filename = sys.argv[1]

    # the mode parameter is passed directly to C's fopen
    # opening the file for writing requires r+, not rw because rw would
    # truncate (effectively destroy) the file, and r+ would preserve the size
    with segyio.open( filename, "r+" ) as src:

        # read trace 0, then double all values
        trace = src.trace[0]
        trace *= 2

        # write trace 0 back to disk
        src.trace[0] = trace

        # read trace 1, but re-use the memory for speed
        trace = src.trace[1, trace]
        # square all values. the trace is a plain numpy array
        trace = np.square(trace, trace)
        # write the trace back to disk, but at trace 2
        src.trace[2] = trace

        # read every other trace, from 10 through 20
        # then write them to every third step from 40 through 52
        # i.e. 40, 43, 46...
        # slices yield a generator, so only one numpy array is created
        for tr, i in zip(src.trace[10:20:2], range(2,13,3)):
            src.trace[i] = tr

        # iterate over all traces in a file. this is a generator with a shared
        # buffer, so it's quite efficient
        tracesum = 0
        for tr in src.trace:
            # accumulate the traces' 30th value
            tracesum += tr[30]

        print("Trace sum: %f" % tracesum)

        # write the iline at 2 to the iline at 3
        sum3 = np.sum(src.iline[3])
        src.iline[2] = src.iline[3]
        # flush to make sure our changes to the file are visible
        src.flush()
        sum2 = np.sum(src.iline[2])

        print("Accumulates of inlines 2 and 3: %f -- %f" % (sum2, sum3))

        # ilines too are plain numpy ndarrays, with trace-major addressing
        # i.e. iline[2,40] would be yield trace#2's 40th value
        iline = src.iline[2]
        # since ilines are numpy arrays they also support numpy operations
        iline = np.add(iline, src.iline[4])

        # lines too have generator support, so we accumulate the 2nd trace's
        # 22nd value.
        linesum = 0
        for line in src.iline:
            linesum += line[2,22]

        print("Inline sum: %f" % linesum)

        # xline access is identical to iline access
        linesum = 0
        for line in src.xline:
            linesum += line[2,22]

        print("Crossline sum: %f" % linesum)

        # accessing a non-existing inline will raise a KeyError
        try:
            src.iline[5000]
            sys.exit("Was able to access non-existing inline")
        except KeyError as e:
            print(str(e))

if __name__ == '__main__':
    main()



from __future__ import absolute_import

import unittest

import numpy
import segyio
import segyio._segyio as _segyio
from .test_context import TestContext


class _segyioTests(unittest.TestCase):
    def setUp(self):
        self.filename = "test-data/small.sgy"

        lines = {
            1: "DATE: 2016-09-19",
            2: "AN INCREASE IN AMPLITUDE EQUALS AN INCREASE IN ACOUSTIC IMPEDANCE",
            3: "Written by libsegyio (python)",
            11: "TRACE HEADER POSITION:",
            12: "  INLINE BYTES 189-193    | OFFSET BYTES 037-041",
            13: "  CROSSLINE BYTES 193-197 |",
            15: "END EBCDIC HEADER"
        }

        rows = segyio.create_text_header(lines)
        rows = bytearray(rows, 'ascii')  # mutable array of bytes
        rows[-1] = 128  # \x80
        self.ACTUAL_TEXT_HEADER = bytes(rows)

    def test_binary_header_size(self):
        self.assertEqual(400, _segyio.binheader_size())

    def test_textheader_size(self):
        self.assertEqual(3200, _segyio.textheader_size())

    def test_open_non_existing_file(self):
        with self.assertRaises(IOError):
            f = _segyio.open("non-existing", "r")

    def test_close_non_existing_file(self):
        with self.assertRaises(TypeError):
            _segyio.close(None)

    def test_open_and_close_file(self):
        f = _segyio.open(self.filename, "r")
        _segyio.close(f)

    def test_open_flush_and_close_file(self):
        _segyio.flush(None)
        f = _segyio.open(self.filename, "r")
        _segyio.flush(f)
        _segyio.close(f)

        # This does not fail for some reason.
        # with self.assertRaises(IOError):
        #     _segyio.flush(f)

    def test_read_text_header(self):
        f = _segyio.open(self.filename, "r")

        self.assertEqual(_segyio.read_textheader(f, 0), self.ACTUAL_TEXT_HEADER)

        with self.assertRaises(Exception):
            _segyio.read_texthdr(None, 0)

        _segyio.close(f)

    def test_write_text_header(self):
        with TestContext("write_text_header") as context:
            context.copy_file(self.filename)

            f = _segyio.open("small.sgy", "r+")

            with self.assertRaises(ValueError):
                _segyio.write_textheader(f, 0, "")

            self.assertEqual(_segyio.read_textheader(f, 0), self.ACTUAL_TEXT_HEADER)

            _segyio.write_textheader(f, 0, "yolo" * 800)

            textheader = _segyio.read_textheader(f, 0)
            textheader = textheader.decode('ascii')  # Because in Python 3.5 bytes are not comparable to strings
            self.assertEqual(textheader, "yolo" * 800)

            _segyio.close(f)

    def test_read_and_write_binary_header(self):
        with self.assertRaises(Exception):
            hdr = _segyio.read_binaryheader(None)

        with self.assertRaises(Exception):
            _segyio.write_binaryheader(None, None)

        with TestContext("read_and_write_bin_header") as context:
            context.copy_file(self.filename)

            f = _segyio.open("small.sgy", "r+")

            binary_header = _segyio.read_binaryheader(f)

            with self.assertRaises(Exception):
                _segyio.write_binaryheader(f, "Not the correct type")

            _segyio.write_binaryheader(f, binary_header)
            _segyio.close(f)

    def test_read_binary_header_fields(self):
        f = _segyio.open(self.filename, "r")

        binary_header = _segyio.read_binaryheader(f)

        with self.assertRaises(TypeError):
            value = _segyio.get_field("s", 0)

        with self.assertRaises(IndexError):
            value = _segyio.get_field(binary_header, -1)

        self.assertEqual(_segyio.get_field(binary_header, 3225), 1)
        self.assertEqual(_segyio.get_field(binary_header, 3221), 50)

        _segyio.close(f)

    def test_line_metrics(self):
        f = _segyio.open(self.filename, "r")

        binary_header = _segyio.read_binaryheader(f)
        ilb = 189
        xlb = 193
        metrics = _segyio.init_metrics(f, binary_header)
        metrics.update(_segyio.init_cube_metrics(f, ilb, xlb,
                                                 metrics['trace_count'],
                                                 metrics['trace0'],
                                                 metrics['trace_bsize']))
        _segyio.close(f)

        sorting = metrics['sorting']
        trace_count = metrics['trace_count']
        inline_count = metrics['iline_count']
        crossline_count = metrics['xline_count']
        offset_count = metrics['offset_count']

        metrics = _segyio.init_line_metrics(sorting, trace_count, inline_count, crossline_count, offset_count)

        self.assertEqual(metrics['xline_length'], 5)
        self.assertEqual(metrics['xline_stride'], 5)
        self.assertEqual(metrics['iline_length'], 5)
        self.assertEqual(metrics['iline_stride'], 1)

        # (sorting, trace_count, inline_count, crossline_count, offset_count)
        metrics = _segyio.init_line_metrics(1, 15, 3, 5, 1)

        self.assertEqual(metrics['xline_length'], 3)
        self.assertEqual(metrics['xline_stride'], 1)
        self.assertEqual(metrics['iline_length'], 5)
        self.assertEqual(metrics['iline_stride'], 3)

        metrics = _segyio.init_line_metrics(2, 15, 3, 5, 1)

        self.assertEqual(metrics['xline_length'], 3)
        self.assertEqual(metrics['xline_stride'], 5)
        self.assertEqual(metrics['iline_length'], 5)
        self.assertEqual(metrics['iline_stride'], 1)

    def test_metrics(self):
        f = _segyio.open(self.filename, "r")
        binary_header = _segyio.read_binaryheader(f)
        ilb = 189
        xlb = 193

        with self.assertRaises(TypeError):
            metrics = _segyio.init_metrics("?", binary_header)

        with self.assertRaises(TypeError):
            metrics = _segyio.init_metrics(f, "?")

        with self.assertRaises(IndexError):
            metrics = _segyio.init_metrics(f, binary_header)
            metrics.update(_segyio.init_cube_metrics(f, ilb + 1, xlb,
                                                     metrics['trace_count'],
                                                     metrics['trace0'],
                                                     metrics['trace_bsize']))

        metrics = _segyio.init_metrics(f, binary_header)
        metrics.update(_segyio.init_cube_metrics(f, ilb, xlb,
                                                 metrics['trace_count'],
                                                 metrics['trace0'],
                                                 metrics['trace_bsize']))

        self.assertEqual(metrics['trace0'], _segyio.textheader_size() + _segyio.binheader_size())
        self.assertEqual(metrics['sample_count'], 50)
        self.assertEqual(metrics['format'], 1)
        self.assertEqual(metrics['trace_bsize'], 200)
        self.assertEqual(metrics['sorting'], 2) # inline sorting = 2, crossline sorting = 1
        self.assertEqual(metrics['trace_count'], 25)
        self.assertEqual(metrics['offset_count'], 1)
        self.assertEqual(metrics['iline_count'], 5)
        self.assertEqual(metrics['xline_count'], 5)

        _segyio.close(f)

        with self.assertRaises(IOError):
            metrics = _segyio.init_metrics(f, binary_header)

    def test_indices(self):
        f = _segyio.open(self.filename, "r")

        binary_header = _segyio.read_binaryheader(f)
        ilb = 189
        xlb = 193
        metrics = _segyio.init_metrics(f, binary_header)
        dmy = numpy.zeros(2, dtype=numpy.intc)

        dummy_metrics = {'xline_count':   2,
                         'iline_count':   2,
                         'offset_count':  1}

        with self.assertRaises(TypeError):
            _segyio.init_indices(".", {}, dmy, dmy, dmy)

        with self.assertRaises(TypeError):
            _segyio.init_indices(f, "-", dmy, dmy, dmy)

        # with self.assertRaises(KeyError):
        #     _segyio.init_indices(f, {}, dmy, dmy, dmy)

        with self.assertRaises(TypeError):
            _segyio.init_indices(f, dummy_metrics, 1, dmy, dmy)

        with self.assertRaises(TypeError):
            _segyio.init_indices(f, dummy_metrics, dmy, 2, dmy)

        with self.assertRaises(TypeError):
            _segyio.init_indices(f, dummy_metrics, dmy, dmy, 2)

        with self.assertRaises(TypeError):
            fdmy = numpy.zeros(1, dtype=numpy.single)
            _segyio.init_indices(f, dummy_metrics, fdmy, dmy, dmy)

        one = numpy.zeros(1, dtype=numpy.intc)
        two = numpy.zeros(2, dtype=numpy.intc)
        off = numpy.zeros(1, dtype=numpy.intc)
        with self.assertRaises(ValueError):
            _segyio.init_indices(f, dummy_metrics, one, two, off)

        with self.assertRaises(ValueError):
            _segyio.init_indices(f, dummy_metrics, two, one, off)

        metrics.update(_segyio.init_cube_metrics(f, ilb, xlb,
                                                 metrics['trace_count'],
                                                 metrics['trace0'],
                                                 metrics['trace_bsize']))

        # Happy Path
        iline_indexes = numpy.zeros(metrics['iline_count'], dtype=numpy.intc)
        xline_indexes = numpy.zeros(metrics['xline_count'], dtype=numpy.intc)
        offsets       = numpy.zeros(metrics['offset_count'], dtype=numpy.intc)
        _segyio.init_indices(f, metrics, iline_indexes, xline_indexes, offsets)

        self.assertListEqual([1, 2, 3, 4, 5], list(iline_indexes))
        self.assertListEqual([20, 21, 22, 23, 24], list(xline_indexes))
        self.assertListEqual([1], list(offsets))

        _segyio.close(f)

    def test_fread_trace0(self):
        f = _segyio.open(self.filename, "r")

        binary_header = _segyio.read_binaryheader(f)
        ilb = 189
        xlb = 193

        metrics = _segyio.init_metrics(f, binary_header)
        metrics.update(_segyio.init_cube_metrics(f, ilb, xlb,
                                                 metrics['trace_count'],
                                                 metrics['trace0'],
                                                 metrics['trace_bsize']))

        sorting = metrics['sorting']
        trace_count = metrics['trace_count']
        inline_count = metrics['iline_count']
        crossline_count = metrics['xline_count']
        offset_count = metrics['offset_count']

        line_metrics = _segyio.init_line_metrics(sorting, trace_count, inline_count, crossline_count, offset_count)

        iline_indexes = numpy.zeros(metrics['iline_count'], dtype=numpy.intc)
        xline_indexes = numpy.zeros(metrics['xline_count'], dtype=numpy.intc)
        offsets       = numpy.zeros(metrics['offset_count'], dtype=numpy.intc)
        _segyio.init_indices(f, metrics, iline_indexes, xline_indexes, offsets)

        with self.assertRaises(KeyError):
            _segyio.fread_trace0(0, len(xline_indexes), line_metrics['iline_stride'], offset_count, iline_indexes, "inline")

        with self.assertRaises(KeyError):
            _segyio.fread_trace0(2, len(iline_indexes), line_metrics['xline_stride'], offset_count, xline_indexes, "crossline")

        value = _segyio.fread_trace0(1, len(xline_indexes), line_metrics['iline_stride'], offset_count, iline_indexes, "inline")
        self.assertEqual(value, 0)

        value = _segyio.fread_trace0(2, len(xline_indexes), line_metrics['iline_stride'], offset_count, iline_indexes, "inline")
        self.assertEqual(value, 5)

        value = _segyio.fread_trace0(21, len(iline_indexes), line_metrics['xline_stride'], offset_count, xline_indexes, "crossline")
        self.assertEqual(value, 1)

        value = _segyio.fread_trace0(22, len(iline_indexes), line_metrics['xline_stride'], offset_count, xline_indexes, "crossline")
        self.assertEqual(value, 2)

        _segyio.close(f)

    def test_get_and_set_field(self):
        hdr = _segyio.empty_traceheader()

        with self.assertRaises(TypeError):
            _segyio.get_field(".", 0)

        with self.assertRaises(TypeError):
            _segyio.set_field(".", 0, 1)

        with self.assertRaises(IndexError):
            _segyio.get_field(hdr, 0)

        with self.assertRaises(IndexError):
            _segyio.set_field(hdr, 0, 1)

        _segyio.set_field(hdr, 1, 127)
        _segyio.set_field(hdr, 5, 67)
        _segyio.set_field(hdr, 9, 19)

        self.assertEqual(_segyio.get_field(hdr, 1), 127)
        self.assertEqual(_segyio.get_field(hdr, 5), 67)
        self.assertEqual(_segyio.get_field(hdr, 9), 19)

    def test_read_and_write_traceheader(self):
        with TestContext("read_and_write_trace_header") as context:
            context.copy_file(self.filename)

            f = _segyio.open("small.sgy", "r+")
            binary_header = _segyio.read_binaryheader(f)
            ilb = 189
            xlb = 193
            metrics = _segyio.init_metrics(f, binary_header)

            empty = _segyio.empty_traceheader()

            with self.assertRaises(TypeError):
                trace_header = _segyio.read_traceheader("+", )

            with self.assertRaises(TypeError):
                trace_header = _segyio.read_traceheader(f, 0, None)

            trace_header = _segyio.read_traceheader(f, 0, _segyio.empty_traceheader(), metrics['trace0'], metrics['trace_bsize'])

            self.assertEqual(_segyio.get_field(trace_header, ilb), 1)
            self.assertEqual(_segyio.get_field(trace_header, xlb), 20)

            trace_header = _segyio.read_traceheader(f, 1, _segyio.empty_traceheader(), metrics['trace0'], metrics['trace_bsize'])

            self.assertEqual(_segyio.get_field(trace_header, ilb), 1)
            self.assertEqual(_segyio.get_field(trace_header, xlb), 21)

            _segyio.set_field(trace_header, ilb, 99)
            _segyio.set_field(trace_header, xlb, 42)

            _segyio.write_traceheader(f, 0, trace_header, metrics['trace0'], metrics['trace_bsize'])

            trace_header = _segyio.read_traceheader(f, 0, _segyio.empty_traceheader(), metrics['trace0'], metrics['trace_bsize'])

            self.assertEqual(_segyio.get_field(trace_header, ilb), 99)
            self.assertEqual(_segyio.get_field(trace_header, xlb), 42)

            _segyio.close(f)

    def test_read_and_write_trace(self):
        with TestContext("read_and_write_trace") as context:
            f = _segyio.open("trace-wrt.sgy", "w+")

            buf = numpy.ones(25, dtype=numpy.single)
            buf[11] = 3.1415
            _segyio.write_trace(f, 0, buf, 0, 100, 1, 25)
            buf[:] = 42.0
            _segyio.write_trace(f, 1, buf, 0, 100, 1, 25)

            _segyio.flush(f)

            buf = numpy.zeros(25, dtype=numpy.single)

            _segyio.read_trace(f, buf, 0, 1, 1, 1, 25, 0, 100)

            self.assertAlmostEqual(buf[10], 1.0, places=4)
            self.assertAlmostEqual(buf[11], 3.1415, places=4)

            _segyio.read_trace(f, buf, 1, 1, 1, 1, 25, 0, 100)

            self.assertAlmostEqual(sum(buf), 42.0 * 25, places=4)

            _segyio.close(f)

    def read_small(self, mmap = False):
        f = _segyio.open(self.filename, "r")

        if mmap: _segyio.mmap(f)

        binary_header = _segyio.read_binaryheader(f)
        ilb = 189
        xlb = 193

        metrics = _segyio.init_metrics(f, binary_header)
        metrics.update(_segyio.init_cube_metrics(f, ilb, xlb,
                                                 metrics['trace_count'],
                                                 metrics['trace0'],
                                                 metrics['trace_bsize']))

        sorting = metrics['sorting']
        trace_count = metrics['trace_count']
        inline_count = metrics['iline_count']
        crossline_count = metrics['xline_count']
        offset_count = metrics['offset_count']

        line_metrics = _segyio.init_line_metrics(sorting, trace_count, inline_count, crossline_count, offset_count)

        metrics.update(line_metrics)

        iline_indexes = numpy.zeros(metrics['iline_count'], dtype=numpy.intc)
        xline_indexes = numpy.zeros(metrics['xline_count'], dtype=numpy.intc)
        offsets       = numpy.zeros(metrics['offset_count'], dtype=numpy.intc)
        _segyio.init_indices(f, metrics, iline_indexes, xline_indexes, offsets)

        return f, metrics, iline_indexes, xline_indexes

    def test_read_line(self):
        f, metrics, iline_idx, xline_idx = self.read_small()

        tr0 = metrics['trace0']
        bsz = metrics['trace_bsize']
        samples = metrics['sample_count']
        xline_stride = metrics['xline_stride']
        iline_stride = metrics['iline_stride']
        offsets = metrics['offset_count']

        xline_trace0 = _segyio.fread_trace0(20, len(iline_idx), xline_stride, offsets, xline_idx, "crossline")
        iline_trace0 = _segyio.fread_trace0(1, len(xline_idx), iline_stride, offsets, iline_idx, "inline")

        buf = numpy.zeros((len(iline_idx), samples), dtype=numpy.single)

        _segyio.read_line(f, xline_trace0, len(iline_idx), xline_stride, offsets, buf, tr0, bsz, 1, samples)
        self.assertAlmostEqual(sum(sum(buf)), 800.061169624, places=6)

        _segyio.read_line(f, iline_trace0, len(xline_idx), iline_stride, offsets, buf, tr0, bsz, 1, samples)
        self.assertAlmostEqual(sum(sum(buf)), 305.061146736, places=6)

        _segyio.close(f)

    def test_read_line_mmap(self):
        f, metrics, iline_idx, xline_idx = self.read_small(True)

        tr0 = metrics['trace0']
        bsz = metrics['trace_bsize']
        samples = metrics['sample_count']
        xline_stride = metrics['xline_stride']
        iline_stride = metrics['iline_stride']
        offsets = metrics['offset_count']

        xline_trace0 = _segyio.fread_trace0(20, len(iline_idx), xline_stride, offsets, xline_idx, "crossline")
        iline_trace0 = _segyio.fread_trace0(1, len(xline_idx), iline_stride, offsets, iline_idx, "inline")

        buf = numpy.zeros((len(iline_idx), samples), dtype=numpy.single)

        _segyio.read_line(f, xline_trace0, len(iline_idx), xline_stride, offsets, buf, tr0, bsz, 1, samples)
        self.assertAlmostEqual(sum(sum(buf)), 800.061169624, places=6)

        _segyio.read_line(f, iline_trace0, len(xline_idx), iline_stride, offsets, buf, tr0, bsz, 1, samples)
        self.assertAlmostEqual(sum(sum(buf)), 305.061146736, places=6)

        _segyio.close(f)

    def test_fread_trace0_for_depth(self):
        elements = list(range(25))
        indices = numpy.asarray(elements, dtype=numpy.intc)

        for index in indices:
            d = _segyio.fread_trace0(index, 1, 1, 1, indices, "depth")
            self.assertEqual(d, index)

        with self.assertRaises(KeyError):
            d = _segyio.fread_trace0(25, 1, 1, 1, indices, "depth")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import os
import sys
from setuptools import setup, Extension

long_description = """
=======
SEGY IO
=======

Introduction
------------

Segyio is a small LGPL licensed C library for easy interaction with SEG Y
formatted seismic data, with language bindings for Python and Matlab. Segyio is
an attempt to create an easy-to-use, embeddable, community-oriented library for
seismic applications. Features are added as they are needed; suggestions and
contributions of all kinds are very welcome.

Feature summary
---------------
 * A low-level C interface with few assumptions; easy to bind to other
   languages.
 * Read and write binary and textual headers.
 * Read and write traces, trace headers.
 * Easy to use and native-feeling python interface with numpy integration.

Project goals
-------------

Segyio does necessarily attempt to be the end-all of SEG-Y interactions;
rather, we aim to lower the barrier to interacting with SEG-Y files for
embedding, new applications or free-standing programs.

Additionally, the aim is not to support the full standard or all exotic (but
correctly) formatted files out there. Some assumptions are made, such as:

 * All traces in a file are assumed to be of the same sample size.
 * It is assumed all lines have the same number of traces.

The writing functionality in Segyio is largely meant to *modify* or adapt
files. A file created from scratch is not necessarily a to-spec SEG-Y file, as
we only necessarily write the header fields segyio needs to make sense of the
geometry. It is still highly recommended that SEG-Y files are maintained and
written according to specification, but segyio does not mandate this.

"""

def src(x):
    root = os.path.dirname( __file__ )
    return os.path.abspath(os.path.join(root, x))

extra_libs = ['m'] if not 'win' in sys.platform else []

def getversion():
    # if this is a tarball distribution, the .git-directory won't be avilable
    # and setuptools_scm will crash hard. good tarballs are built with a
    # pre-generated version.py, so parse that and extract version from it
    if os.path.isdir(src('.git')):
        return {
            'use_scm_version': {
                        'root': src(''),
                        'write_to': src('python/segyio/version.py')
            }
        }


    pkgversion = { 'version': '0.0.0' }
    versionfile = src('python/segyio/version.py')

    if not os.path.exists(versionfile):
        return pkgversion

    import ast
    with open(versionfile) as f:
        root = ast.parse(f.read())

    for node in ast.walk(root):
        if not isinstance(node, ast.Assign): continue
        if len(node.targets) == 1 and node.targets[0].id == 'version':
            pkgversion['version'] = node.value.s

    return pkgversion

setup(name='segyio',
      description='Simple & fast IO for SEG-Y files',
      long_description=long_description,
      author='Statoil ASA',
      author_email='fg_gpl@statoil.com',
      url='https://github.com/Statoil/segyio',
      package_dir={'' : src('python')},
      packages=['segyio'],
      package_data={'': ['License', 'README']},
      license='LGPL-3.0',
      ext_modules=[Extension('segyio._segyio',
        sources=[src('python/segyio/segyio.cpp')],
        include_dirs=[src('lib/include')],
        libraries=['segyio'] + extra_libs
        )],
      platforms='any',
      requires=['numpy'],
      install_requires=['numpy'],
      test_suite='test',
      setup_requires=['setuptools_scm'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Other Environment',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Software Development :: Libraries',
          'Topic :: Utilities'
      ],
      **getversion()
      )



from __future__ import absolute_import

from types import GeneratorType

import itertools
import numpy as np
import unittest
from .test_context import TestContext

import segyio
from segyio import TraceField, BinField
import filecmp

from segyio._field import Field
from segyio._line import Line
from segyio._header import Header
from segyio._trace import Trace

try:
    from itertools import izip as zip
    from itertools import imap as map
except ImportError:  # will be 3.x series
    pass


def mklines(fname):
    spec = segyio.spec()
    spec.format  = 5
    spec.sorting = 2
    spec.samples = range(10)
    spec.ilines  = range(1, 11)
    spec.xlines  = range(1, 6)

    # create a file with 10 inlines, with values on the form l.0tv where
    # l = line no
    # t = trace number (within line)
    # v = trace value
    # i.e. 2.043 is the value at inline 2's fourth trace's third value
    with segyio.create(fname, spec) as dst:
        ln = np.arange(start = 0,
                       stop = 0.001 * (5 * 10),
                       step = 0.001,
                       dtype = np.single).reshape(5, 10)

        for il in spec.ilines:
            ln += 1

            dst.header.iline[il] = { TraceField.INLINE_3D: il }
            dst.iline[il] = ln

        for xl in spec.xlines:
            dst.header.xline[xl] = { TraceField.CROSSLINE_3D: xl }


class TestSegy(unittest.TestCase):

    def setUp(self):
        self.filename = "test-data/small.sgy"
        self.prestack = "test-data/small-ps.sgy"
        self.fileMx1  = "test-data/Mx1.sgy"
        self.file1xN  = "test-data/1xN.sgy"
        self.file1x1  = "test-data/1x1.sgy"

    def test_inline_4(self):
        with segyio.open(self.filename, "r") as f:

            sample_count = len(f.samples)
            self.assertEqual(50, sample_count)

            data = f.iline[4]

            self.assertAlmostEqual(4.2, data[0, 0], places = 6)
            # middle sample
            self.assertAlmostEqual(4.20024, data[0, sample_count//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(4.20049, data[0, -1], places = 6)

            # middle xline
            middle_line = 2
            # first sample
            self.assertAlmostEqual(4.22, data[middle_line, 0], places = 5)
            # middle sample
            self.assertAlmostEqual(4.22024, data[middle_line, sample_count//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(4.22049, data[middle_line, -1], places = 6)

            # last xline
            last_line = (len(f.xlines)-1)
            # first sample
            self.assertAlmostEqual(4.24, data[last_line, 0], places = 5)
            # middle sample
            self.assertAlmostEqual(4.24024, data[last_line, sample_count//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(4.24049, data[last_line, sample_count-1], places = 6)

    def test_xline_22(self):
        with segyio.open(self.filename, "r") as f:

            data = f.xline[22]

            size = len(f.samples)

            # first iline
            # first sample
            self.assertAlmostEqual(1.22, data[0, 0], places = 5)
            # middle sample
            self.assertAlmostEqual(1.22024, data[0, size//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(1.22049, data[0, size-1], places = 6)

            # middle iline
            middle_line = 2
            # first sample
            self.assertAlmostEqual(3.22, data[middle_line, 0], places = 5)
            # middle sample
            self.assertAlmostEqual(3.22024, data[middle_line, size//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(3.22049, data[middle_line, size-1], places = 6)

            # last iline
            last_line = len(f.ilines)-1
            # first sample
            self.assertAlmostEqual(5.22, data[last_line, 0], places = 5)
            # middle sample
            self.assertAlmostEqual(5.22024, data[last_line, size//2-1], places = 6)
            # last sample
            self.assertAlmostEqual(5.22049, data[last_line, size-1], places = 6)

    def test_iline_slicing(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(len(f.ilines), sum(1 for _ in f.iline))
            self.assertEqual(len(f.ilines), sum(1 for _ in f.iline[1:6]))
            self.assertEqual(len(f.ilines), sum(1 for _ in f.iline[5:0:-1]))
            self.assertEqual(len(f.ilines) // 2, sum(1 for _ in f.iline[0::2]))
            self.assertEqual(len(f.ilines), sum(1 for _ in f.iline[1:]))
            self.assertEqual(3,  sum(1 for _ in f.iline[::2]))
            self.assertEqual(0,  sum(1 for _ in f.iline[12:24]))
            self.assertEqual(3,  sum(1 for _ in f.iline[:4]))
            self.assertEqual(2,  sum(1 for _ in f.iline[2:6:2]))

    def test_xline_slicing(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(len(f.xlines), sum(1 for _ in f.xline))
            self.assertEqual(len(f.xlines), sum(1 for _ in f.xline[20:25]))
            self.assertEqual(len(f.xlines), sum(1 for _ in f.xline[25:19:-1]))
            self.assertEqual(3, sum(1 for _ in f.xline[0::2]))
            self.assertEqual(3, sum(1 for _ in f.xline[::2]))
            self.assertEqual(len(f.xlines), sum(1 for _ in f.xline[20:]))
            self.assertEqual(0,  sum(1 for _ in f.xline[12:18]))
            self.assertEqual(5,  sum(1 for _ in f.xline[:25]))
            self.assertEqual(2,  sum(1 for _ in f.xline[:25:3]))

    def test_open_transposed_lines(self):
        il, xl = [], []
        with segyio.open(self.filename, "r") as f:
            il = f.ilines
            xl = f.xlines

        with segyio.open(self.filename, "r", segyio.TraceField.CROSSLINE_3D, segyio.TraceField.INLINE_3D) as f:
            self.assertListEqual(list(il), list(f.xlines))
            self.assertListEqual(list(xl), list(f.ilines))
            pass

    def test_file_info(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(2, f.sorting)
            self.assertEqual(1, f.offsets)
            self.assertEqual(1, int(f.format))

            xlines = list(range(20, 25))
            ilines = list(range(1, 6))
            self.assertEqual(xlines, list(f.xlines))
            self.assertEqual(ilines, list(f.ilines))
            self.assertEqual(25, f.tracecount)
            self.assertEqual(len(f.trace), f.tracecount)
            self.assertEqual(50, len(f.samples))

    def test_open_nostrict(self):
        with segyio.open(self.filename, strict = False):
            pass

    def test_open_ignore_geometry(self):
        with segyio.open(self.filename, ignore_geometry = True) as f:
            with self.assertRaises(ValueError):
                f.iline[0]

    def test_traces_slicing(self):
        with segyio.open(self.filename, "r") as f:

            traces = list(map(np.copy, f.trace[0:6:2]))
            self.assertEqual(len(traces), 3)
            self.assertEqual(traces[0][49], f.trace[0][49])
            self.assertEqual(traces[1][49], f.trace[2][49])
            self.assertEqual(traces[2][49], f.trace[4][49])

            rev_traces = list(map(np.copy, f.trace[4::-2]))
            self.assertEqual(rev_traces[0][49], f.trace[4][49])
            self.assertEqual(rev_traces[1][49], f.trace[2][49])
            self.assertEqual(rev_traces[2][49], f.trace[0][49])

            # make sure buffers can be reused
            buf = None
            for i, trace in enumerate(f.trace[0:6:2, buf]):
                self.assertTrue(np.array_equal(trace, traces[i]))

    def test_traces_offset(self):
        with segyio.open(self.prestack, "r") as f:

            self.assertEqual(2, len(f.offsets))
            self.assertListEqual([1, 2], list(f.offsets))

            # traces are laid out |l1o1 l1o2 l2o1 l2o2...|
            # where l = iline number and o = offset number
            # traces are not re-indexed according to offsets
            # see make-ps-file.py for value formula
            self.assertAlmostEqual(101.01, f.trace[0][0], places = 4)
            self.assertAlmostEqual(201.01, f.trace[1][0], places = 4)
            self.assertAlmostEqual(101.02, f.trace[2][0], places = 4)
            self.assertAlmostEqual(201.02, f.trace[3][0], places = 4)
            self.assertAlmostEqual(102.01, f.trace[6][0], places = 4)

    def test_headers_offset(self):
        with segyio.open(self.prestack, "r") as f:
            il, xl = TraceField.INLINE_3D, TraceField.CROSSLINE_3D
            self.assertEqual(f.header[0][il], f.header[1][il])
            self.assertEqual(f.header[1][il], f.header[2][il])

            self.assertEqual(f.header[0][xl], f.header[1][xl])
            self.assertNotEqual(f.header[1][xl], f.header[2][xl])

    def test_header_dict_methods(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(89, len(list(f.header[0].keys())))
            self.assertEqual(89, len(list(f.header[1].values())))
            self.assertEqual(89, len(list(f.header[2].items())))
            self.assertEqual(89, len(list(f.header[3])))
            self.assertTrue(0 not in f.header[0])
            self.assertTrue(1 in f.header[0])
            self.assertTrue(segyio.su.cdpx in f.header[0])
            iter(f.header[0])

            self.assertEqual(30, len(f.bin.keys()))
            self.assertEqual(30, len(list(f.bin.values())))
            self.assertEqual(30, len(list(f.bin.items())))
            self.assertEqual(30, len(f.bin))
            iter(f.bin)

    def test_headers_line_offset(self):
        with TestContext("headers_line_offset") as context:
            context.copy_file(self.prestack)
            il, xl = TraceField.INLINE_3D, TraceField.CROSSLINE_3D
            with segyio.open("small-ps.sgy", "r+") as f:
                f.header.iline[1,2] = { il: 11 }
                f.header.iline[1,2] = { xl: 13 }

            with segyio.open("small-ps.sgy", "r", strict = False) as f:
                self.assertEqual(f.header[0][il], 1)
                self.assertEqual(f.header[1][il], 11)
                self.assertEqual(f.header[2][il], 1)

                self.assertEqual(f.header[0][xl], 1)
                self.assertEqual(f.header[1][xl], 13)
                self.assertEqual(f.header[2][xl], 2)

    def test_attributes(self):
        with segyio.open(self.filename) as f:
            il = TraceField.INLINE_3D
            xl = TraceField.CROSSLINE_3D

            self.assertEqual(1,  f.attributes(il)[0])
            self.assertEqual(20, f.attributes(xl)[0])

            ils = [(i // 5) + 1 for i in range(25)]
            attrils = list(map(int, f.attributes(il)[:]))
            self.assertListEqual(ils, attrils)

            xls = [(i % 5) + 20 for i in range(25)]
            attrxls = list(map(int, f.attributes(xl)[:]))
            self.assertListEqual(xls, attrxls)

            ils = [(i // 5) + 1 for i in range(25)][::-1]
            attrils = list(map(int, f.attributes(il)[::-1]))
            self.assertListEqual(ils, attrils)

            xls = [(i % 5) + 20 for i in range(25)][::-1]
            attrxls = list(map(int, f.attributes(xl)[::-1]))
            self.assertListEqual(xls, attrxls)

            self.assertEqual(f.header[0][il], f.attributes(il)[0])
            f.mmap()
            self.assertEqual(f.header[0][il], f.attributes(il)[0])

            ils = [(i // 5) + 1 for i in range(25)][1:21:3]
            attrils = list(map(int, f.attributes(il)[1:21:3]))
            self.assertListEqual(ils, attrils)

            xls = [(i % 5) + 20 for i in range(25)][2:17:5]
            attrxls = list(map(int, f.attributes(xl)[2:17:5]))
            self.assertListEqual(xls, attrxls)

            ils = [1, 2, 3, 4, 5]
            attrils = list(map(int, f.attributes(il)[[0, 5, 11, 17, 23]]))
            self.assertListEqual(ils, attrils)

            ils = [1, 2, 3, 4, 5]
            indices = np.asarray([0, 5, 11, 17, 23])
            attrils = list(map(int, f.attributes(il)[indices]))
            self.assertListEqual(ils, attrils)

    def test_iline_offset(self):
        with segyio.open(self.prestack, "r") as f:

            line1 = f.iline[1, 1]
            self.assertAlmostEqual(101.01, line1[0][0], places = 4)
            self.assertAlmostEqual(101.02, line1[1][0], places = 4)
            self.assertAlmostEqual(101.03, line1[2][0], places = 4)

            self.assertAlmostEqual(101.01001, line1[0][1], places = 4)
            self.assertAlmostEqual(101.01002, line1[0][2], places = 4)
            self.assertAlmostEqual(101.02001, line1[1][1], places = 4)

            line2 = f.iline[1, 2]
            self.assertAlmostEqual(201.01, line2[0][0], places = 4)
            self.assertAlmostEqual(201.02, line2[1][0], places = 4)
            self.assertAlmostEqual(201.03, line2[2][0], places = 4)

            self.assertAlmostEqual(201.01001, line2[0][1], places = 4)
            self.assertAlmostEqual(201.01002, line2[0][2], places = 4)
            self.assertAlmostEqual(201.02001, line2[1][1], places = 4)

            with self.assertRaises(KeyError):
                f.iline[1, 0]

            with self.assertRaises(KeyError):
                f.iline[1, 3]

            with self.assertRaises(KeyError):
                f.iline[100, 1]

            with self.assertRaises(TypeError):
                f.iline[1, {}]

    def test_iline_slice_fixed_offset(self):
        with segyio.open(self.prestack, "r") as f:

            for i, ln in enumerate(f.iline[:, 1], 1):
                self.assertAlmostEqual(i + 100.01, ln[0][0], places = 4)
                self.assertAlmostEqual(i + 100.02, ln[1][0], places = 4)
                self.assertAlmostEqual(i + 100.03, ln[2][0], places = 4)

                self.assertAlmostEqual(i + 100.01001, ln[0][1], places = 4)
                self.assertAlmostEqual(i + 100.01002, ln[0][2], places = 4)
                self.assertAlmostEqual(i + 100.02001, ln[1][1], places = 4)

    def test_iline_slice_fixed_line(self):
        with segyio.open(self.prestack, "r") as f:

            for i, ln in enumerate(f.iline[1, :], 1):
                off = i * 100
                self.assertAlmostEqual(off + 1.01, ln[0][0], places = 4)
                self.assertAlmostEqual(off + 1.02, ln[1][0], places = 4)
                self.assertAlmostEqual(off + 1.03, ln[2][0], places = 4)

                self.assertAlmostEqual(off + 1.01001, ln[0][1], places = 4)
                self.assertAlmostEqual(off + 1.01002, ln[0][2], places = 4)
                self.assertAlmostEqual(off + 1.02001, ln[1][1], places = 4)

    def test_iline_slice_all_offsets(self):
        with segyio.open(self.prestack, "r") as f:
            offs, ils = len(f.offsets), len(f.ilines)
            self.assertEqual(offs * ils, sum(1 for _ in f.iline[:,:]))
            self.assertEqual(offs * ils, sum(1 for _ in f.iline[:,::-1]))
            self.assertEqual(offs * ils, sum(1 for _ in f.iline[::-1,:]))
            self.assertEqual(offs * ils, sum(1 for _ in f.iline[::-1,::-1]))
            self.assertEqual(0,  sum(1 for _ in f.iline[:, 10:12]))
            self.assertEqual(0,  sum(1 for _ in f.iline[10:12, :]))

            self.assertEqual((offs // 2) * ils, sum(1 for _ in f.iline[::2, :]))
            self.assertEqual(offs * (ils // 2), sum(1 for _ in f.iline[:, ::2]))

            self.assertEqual((offs // 2) * ils, sum(1 for _ in f.iline[::-2, :]))
            self.assertEqual(offs * (ils // 2), sum(1 for _ in f.iline[:, ::-2]))

            self.assertEqual((offs // 2) * (ils // 2), sum(1 for _ in f.iline[::2, ::2]))
            self.assertEqual((offs // 2) * (ils // 2), sum(1 for _ in f.iline[::2, ::-2]))
            self.assertEqual((offs // 2) * (ils // 2), sum(1 for _ in f.iline[::-2, ::2]))


    def test_gather_mode(self):
        with segyio.open(self.prestack) as f:
            empty = np.empty(0, dtype = np.single)
            # should raise
            with self.assertRaises(KeyError):
                self.assertTrue(np.array_equal(empty, f.gather[2, 3, 3]))

            with self.assertRaises(KeyError):
                self.assertTrue(np.array_equal(empty, f.gather[2, 5, 1]))

            with self.assertRaises(KeyError):
                self.assertTrue(np.array_equal(empty, f.gather[5, 2, 1]))

            self.assertTrue(np.array_equal(f.trace[10], f.gather[2, 3, 1]))
            self.assertTrue(np.array_equal(f.trace[11], f.gather[2, 3, 2]))

            traces = segyio.tools.collect(f.trace[10:12])
            gather = f.gather[2,3,:]
            self.assertTrue(np.array_equal(traces, gather))
            self.assertTrue(np.array_equal(traces, f.gather[2,3]))
            self.assertTrue(np.array_equal(empty,  f.gather[2,3,1:0]))
            self.assertTrue(np.array_equal(empty,  f.gather[2,3,3:4]))

            for g, line in zip(f.gather[1:3, 3, 1], f.iline[1:3]):
                self.assertEqual(10, len(g))
                self.assertEqual((10,), g.shape)
                self.assertTrue(np.array_equal(line[2], g))

            for g, line in zip(f.gather[1:3, 3, :], f.iline[1:3]):
                self.assertEqual(2, len(g))
                self.assertEqual((2, 10), g.shape)
                self.assertTrue(np.array_equal(line[2], g[0]))

            for g, line in zip(f.gather[1, 1:3, 1], f.xline[1:3]):
                self.assertEqual(10, len(g))
                self.assertEqual((10,), g.shape)
                self.assertTrue(np.array_equal(line[0], g))

            for g, line in zip(f.gather[1, 1:3, :], f.xline[1:3]):
                self.assertEqual(2, len(g))
                self.assertEqual((2, 10), g.shape)
                self.assertTrue(np.array_equal(line[0], g[0]))


    def test_line_generators(self):
        with segyio.open(self.filename, "r") as f:
            for line in f.iline:
                pass

            for line in f.xline:
                pass

    def test_fast_slow_dimensions(self):
        with segyio.open(self.filename, 'r') as f:
            for iline, fline in zip(f.iline, f.fast):
                self.assertTrue(np.array_equal(iline, fline))

            for xline, sline in zip(f.xline, f.slow):
                self.assertTrue(np.array_equal(iline, fline))

    def test_traces_raw(self):
        with segyio.open(self.filename, "r") as f:
            gen_traces = np.array(list(map( np.copy, f.trace )), dtype = np.single)

            raw_traces = f.trace.raw[:]
            self.assertTrue(np.array_equal(gen_traces, raw_traces))

            self.assertEqual(len(gen_traces), f.tracecount)
            self.assertEqual(len(raw_traces), f.tracecount)

            self.assertEqual(gen_traces[0][49], raw_traces[0][49])
            self.assertEqual(gen_traces[1][49], f.trace.raw[1][49])
            self.assertEqual(gen_traces[2][49], raw_traces[2][49])

            self.assertTrue(np.array_equal(f.trace[10], f.trace.raw[10]))

            for raw, gen in zip(f.trace.raw[::2], f.trace[::2]):
                self.assertTrue(np.array_equal(raw, gen))

            for raw, gen in zip(f.trace.raw[::-1], f.trace[::-1]):
                self.assertTrue(np.array_equal(raw, gen))

    def test_read_header(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(1, f.header[0][189])
            self.assertEqual(1, f.header[1][TraceField.INLINE_3D])
            self.assertEqual(1, f.header[1][segyio.su.iline])
            self.assertEqual(5, f.header[-1][segyio.su.iline])
            self.assertEqual(5, f.header[24][segyio.su.iline])
            self.assertEqual(dict(f.header[-1]), dict(f.header[24]))

            with self.assertRaises(IndexError):
                f.header[30]

            with self.assertRaises(IndexError):
                f.header[-30]

            with self.assertRaises(IndexError):
                f.header[0][188] # between byte offsets

            with self.assertRaises(IndexError):
                f.header[0][-1]

            with self.assertRaises(IndexError):
                f.header[0][700]

    def test_write_header(self):
        with TestContext("write_header") as context:
            context.copy_file(self.filename)
            with segyio.open("small.sgy", "r+") as f:
                # assign to a field in a header, write immediately
                f.header[0][189] = 42
                f.flush()

                self.assertEqual(42, f.header[0][189])
                self.assertEqual(1, f.header[1][189])

                # accessing non-existing offsets raises exceptions
                with self.assertRaises(IndexError):
                    f.header[0][188] = 1 # between byte offsets

                with self.assertRaises(IndexError):
                    f.header[0][-1] = 1

                with self.assertRaises(IndexError):
                    f.header[0][700] = 1

                d = { TraceField.INLINE_3D: 43,
                      TraceField.CROSSLINE_3D: 11,
                      TraceField.offset: 15 }

                # assign multiple fields at once by using a dict
                f.header[1] = d

                f.flush()
                self.assertEqual(43, f.header[1][TraceField.INLINE_3D])
                self.assertEqual(11, f.header[1][segyio.su.xline])
                self.assertEqual(15, f.header[1][segyio.su.offset])

                # looking up multiple values at once returns a { TraceField: value } dict
                self.assertEqual(d, f.header[1][TraceField.INLINE_3D, TraceField.CROSSLINE_3D, TraceField.offset])

                # slice-support over headers (similar to trace)
                for th in f.header[0:10]:
                    pass

                self.assertEqual(6, len(list(f.header[10::-2])))
                self.assertEqual(5, len(list(f.header[10:5:-1])))
                self.assertEqual(0, len(list(f.header[10:5])))

                d = { TraceField.INLINE_3D: 45,
                      TraceField.CROSSLINE_3D: 10,
                      TraceField.offset: 16 }

                # assign multiple values using alternative syntax
                f.header[5].update(d)
                f.flush()
                self.assertEqual(45, f.header[5][TraceField.INLINE_3D])
                self.assertEqual(10, f.header[5][segyio.su.xline])
                self.assertEqual(16, f.header[5][segyio.su.offset])

                # accept anything with a key-value structure
                f.header[5].update( [(segyio.su.ns, 12), (segyio.su.dt, 4)] )
                f.header[5].update( ((segyio.su.muts, 3), (segyio.su.mute, 7)) )

                with self.assertRaises(TypeError):
                    f.header[0].update(10)

                with self.assertRaises(TypeError):
                    f.header[0].update(None)

                with self.assertRaises(ValueError):
                    f.header[0].update('foo')

                f.flush()
                self.assertEqual(12, f.header[5][segyio.su.ns])
                self.assertEqual( 4, f.header[5][segyio.su.dt])
                self.assertEqual( 3, f.header[5][segyio.su.muts])
                self.assertEqual( 7, f.header[5][segyio.su.mute])

                # for-each support
                for th in f.header:
                    pass

                # copy a header
                f.header[2] = f.header[1]
                f.flush()

                # don't use this interface in production code, it's only for testing
                # i.e. don't access buf of treat it as a list
                #self.assertEqual(list(f.header[2].buf), list(f.header[1].buf))

    def test_write_binary(self):
        with TestContext("write_binary") as context:
            context.copy_file(self.filename)
            with segyio.open("small.sgy", "r+") as f:
                f.bin[3213] = 5
                f.flush()

                self.assertEqual(5, f.bin[3213])

                # accessing non-existing offsets raises exceptions
                with self.assertRaises(IndexError):
                    f.bin[0]

                with self.assertRaises(IndexError):
                    f.bin[50000]

                with self.assertRaises(IndexError):
                    f.bin[3214]

                d = { BinField.Traces: 43,
                      BinField.SweepFrequencyStart: 11 }

                # assign multiple fields at once by using a dict
                f.bin = d

                f.flush()
                self.assertEqual(43, f.bin[segyio.su.ntrpr])
                self.assertEqual(11, f.bin[segyio.su.hsfs])

                d = { BinField.Traces: 45,
                      BinField.SweepFrequencyStart: 10 }

                # assign multiple values using alternative syntax
                f.bin.update(d)
                f.flush()
                self.assertEqual(45, f.bin[segyio.su.ntrpr])
                self.assertEqual(10, f.bin[segyio.su.hsfs])

                # accept anything with a key-value structure
                f.bin.update( [(segyio.su.jobid, 12), (segyio.su.lino, 4)] )
                f.bin.update( ((segyio.su.reno, 3), (segyio.su.hdt, 7)) )

                f.flush()
                self.assertEqual(12, f.bin[segyio.su.jobid])
                self.assertEqual( 4, f.bin[segyio.su.lino])
                self.assertEqual( 3, f.bin[segyio.su.reno])
                self.assertEqual( 7, f.bin[segyio.su.hdt])

                # looking up multiple values at once returns a { TraceField: value } dict
                self.assertEqual(d, f.bin[BinField.Traces, BinField.SweepFrequencyStart])

                # copy a header
                f.bin = f.bin

    def test_fopen_error(self):
        # non-existent file
        with self.assertRaises(IOError):
            segyio.open("no_dir/no_file", "r")

        # non-existant mode
        with self.assertRaises(IOError):
            segyio.open(self.filename, "foo")

    def test_wrong_lineno(self):
        with self.assertRaises(KeyError):
            with segyio.open(self.filename, "r") as f:
                f.iline[3000]

        with self.assertRaises(KeyError):
            with segyio.open(self.filename, "r") as f:
                f.xline[2]

    def test_open_wrong_inline(self):
        with self.assertRaises(IndexError):
            with segyio.open(self.filename, "r", 2) as f:
                pass

        with segyio.open(self.filename, "r", 2, strict = False) as f:
            pass

    def test_open_wrong_crossline(self):
        with self.assertRaises(IndexError):
            with segyio.open(self.filename, "r", 189, 2) as f:
                pass

        with segyio.open(self.filename, "r", 189, 2, strict = False) as f:
            pass

    def test_wonky_dimensions(self):
        with segyio.open(self.fileMx1) as f:
            pass

        with segyio.open(self.file1xN) as f:
            pass

        with segyio.open(self.file1x1) as f:
            pass

    def test_open_fails_unstructured(self):
        with segyio.open(self.filename, "r", 37, strict = False) as f:
            with self.assertRaises(ValueError):
                f.iline[10]

            with self.assertRaises(ValueError):
                f.iline[:,:]

            with self.assertRaises(ValueError):
                f.xline[:,:]

            with self.assertRaises(ValueError):
                f.depth_slice[2]

            # operations that don't rely on geometry still works
            self.assertEqual(f.header[2][189], 1)
            self.assertListEqual(list(f.attributes(189)[:]),
                                 [(i // 5) + 1 for i in range(len(f.trace))])

    def test_assign_all_traces(self):
        with TestContext("assign_all_traces") as context:
            context.copy_file(self.filename, target = 'dst')
            context.copy_file(self.filename)
            with segyio.open('small.sgy') as f:
                traces = f.trace.raw[:] * 2.0

            with segyio.open('dst/small.sgy', 'r+') as f:
                f.trace[:] = traces[:]

            with segyio.open('dst/small.sgy') as f:
                self.assertTrue(np.array_equal(f.trace.raw[:], traces))

    def test_traceaccess_from_array(self):
        a = np.arange(10, dtype = np.int)
        b = np.arange(10, dtype = np.int32)
        c = np.arange(10, dtype = np.int64)
        d = np.arange(10, dtype = np.intc)
        with segyio.open(self.filename) as f:
            f.trace[a[0]]
            f.trace[b[1]]
            f.trace[c[2]]
            f.trace[d[3]]

    def test_create_sgy(self):
        with TestContext("create_sgy") as context:
            context.copy_file(self.filename)
            src_file = "small.sgy"
            dst_file = "small_created.sgy"
            with segyio.open(src_file, "r") as src:
                spec = segyio.spec()
                spec.format     = int(src.format)
                spec.sorting    = int(src.sorting)
                spec.samples    = src.samples
                spec.ilines     = src.ilines
                spec.xlines     = src.xlines

                with segyio.create(dst_file, spec) as dst:
                    dst.text[0] = src.text[0]
                    dst.bin     = src.bin

                    # copy all headers
                    dst.header = src.header

                    for i, srctr in enumerate(src.trace):
                        dst.trace[i] = srctr

                    dst.trace = src.trace

                    # this doesn't work yet, some restructuring is necessary
                    # if it turns out to be a desired feature it's rather easy to do
                    #for dsth, srch in zip(dst.header, src.header):
                    #    dsth = srch

                    #for dsttr, srctr in zip(dst.trace, src.trace):
                    #    dsttr = srctr

            self.assertTrue(filecmp.cmp(src_file, dst_file))

    def test_create_sgy_shorter_traces(self):
        with TestContext("create_sgy_shorter_traces") as context:
            context.copy_file(self.filename)
            src_file = "small.sgy"
            dst_file = "small_created_shorter.sgy"

            with segyio.open(src_file, "r") as src:
                spec = segyio.spec()
                spec.format  = int(src.format)
                spec.sorting = int(src.sorting)
                spec.samples = src.samples[:20] # reduces samples per trace
                spec.ilines  = src.ilines
                spec.xlines  = src.xlines

                with segyio.create(dst_file, spec) as dst:
                    for i, srch in enumerate(src.header):
                        dst.header[i] = srch
                        d = { TraceField.INLINE_3D: srch[TraceField.INLINE_3D] + 100 }
                        dst.header[i] = d

                    for lineno in dst.ilines:
                        dst.iline[lineno] = src.iline[lineno]

                    # alternative form using left-hand-side slices
                    dst.iline[2:4] = src.iline

                    for lineno in dst.xlines:
                        dst.xline[lineno] = src.xline[lineno]

                with segyio.open(dst_file, "r") as dst:
                    self.assertEqual(20, len(dst.samples))
                    self.assertEqual([x + 100 for x in src.ilines], list(dst.ilines))

    def test_create_from_naught(self):
        with TestContext("create_from_naught") as context:
            fname = "mk.sgy"
            spec = segyio.spec()
            spec.format  = 5
            spec.sorting = 2
            spec.samples = range(150)
            spec.ilines  = range(1, 11)
            spec.xlines  = range(1, 6)

            with segyio.create(fname, spec) as dst:
                tr = np.arange( start = 1.000, stop = 1.151, step = 0.001, dtype = np.single)

                for i in range( len( dst.trace ) ):
                    dst.trace[i] = tr
                    tr += 1.000

                for il in spec.ilines:
                    dst.header.iline[il] = { TraceField.INLINE_3D: il }

                for xl in spec.xlines:
                    dst.header.xline[xl] = { TraceField.CROSSLINE_3D: xl }

                # Set header field 'offset' to 1 in all headers
                dst.header = { TraceField.offset: 1 }

            with segyio.open(fname, "r") as f:
                self.assertAlmostEqual(1,      f.trace[0][0],   places = 4)
                self.assertAlmostEqual(1.001,  f.trace[0][1],   places = 4)
                self.assertAlmostEqual(1.149,  f.trace[0][-1],   places = 4)
                self.assertAlmostEqual(50.100, f.trace[-1][100], places = 4)
                self.assertEqual(f.header[0][TraceField.offset], f.header[1][TraceField.offset])
                self.assertEqual(1, f.header[1][TraceField.offset])

    def test_create_from_naught_prestack(self):
        with TestContext("create_from_naught_prestack") as context:
            fname = "mk-ps.sgy"
            spec = segyio.spec()
            spec.format  = 5
            spec.sorting = 2
            spec.samples = range(7)
            spec.ilines  = range(1, 4)
            spec.xlines  = range(1, 3)
            spec.offsets = range(1, 6)

            cube_size = len(spec.ilines) * len(spec.xlines)

            with segyio.create(fname, spec) as dst:
                arr = np.arange( start = 0.000,
                                 stop = 0.007,
                                 step = 0.001,
                                 dtype = np.single)

                arr = np.concatenate([[arr + 0.01], [arr + 0.02]], axis = 0)
                lines = [arr + i for i in spec.ilines]
                cube = [(off * 100) + line for line in lines for off in spec.offsets]

                dst.iline[:,:] = cube

                for of in spec.offsets:
                    for il in spec.ilines:
                        dst.header.iline[il,of] = { TraceField.INLINE_3D: il,
                                                    TraceField.offset: of
                                                  }
                    for xl in spec.xlines:
                        dst.header.xline[xl,of] = { TraceField.CROSSLINE_3D: xl }

            with segyio.open(fname, "r") as f:
                self.assertAlmostEqual(101.010, f.trace[0][0],  places = 4)
                self.assertAlmostEqual(101.011, f.trace[0][1],  places = 4)
                self.assertAlmostEqual(101.016, f.trace[0][-1], places = 4)
                self.assertAlmostEqual(503.025, f.trace[-1][5], places = 4)
                self.assertNotEqual(f.header[0][TraceField.offset], f.header[1][TraceField.offset])
                self.assertEqual(1, f.header[0][TraceField.offset])
                self.assertEqual(2, f.header[1][TraceField.offset])

                for x, y in zip(f.iline[:,:], cube):
                    self.assertListEqual(list(x.flatten()), list(y.flatten()))

    def test_create_write_lines(self):
        fname = "mklines.sgy"

        with TestContext("create_write_lines") as context:
            mklines(fname)

            with segyio.open(fname, "r") as f:
                self.assertAlmostEqual(1,     f.iline[1][0][0], places = 4)
                self.assertAlmostEqual(2.004, f.iline[2][0][4], places = 4)
                self.assertAlmostEqual(2.014, f.iline[2][1][4], places = 4)
                self.assertAlmostEqual(8.043, f.iline[8][4][3], places = 4)

    def test_create_sgy_skip_lines(self):
        fname = "lines.sgy"
        dstfile = "lines-halved.sgy"

        with TestContext("create_sgy_skip_lines") as context:
            mklines(fname)

            with segyio.open(fname, "r") as src:
                spec = segyio.spec()
                spec.format     = int(src.format)
                spec.sorting    = int(src.sorting)
                spec.samples    = src.samples
                spec.ilines     = src.ilines[::2]
                spec.xlines     = src.xlines[::2]

                with segyio.create(dstfile, spec) as dst:
                    # use the inline headers as base
                    dst.header.iline = src.header.iline[::2]
                    # then update crossline numbers from the crossline headers
                    for xl in dst.xlines:
                        f = next(src.header.xline[xl])[TraceField.CROSSLINE_3D]
                        dst.header.xline[xl] = { TraceField.CROSSLINE_3D: f }

                    # but we override the last xline to be 6, not 5
                    dst.header.xline[5] = { TraceField.CROSSLINE_3D: 6 }
                    dst.iline = src.iline[::2]

            with segyio.open(dstfile, "r") as f:
                self.assertListEqual(list(f.ilines), list(spec.ilines))
                self.assertListEqual(list(f.xlines), [1, 3, 6])
                self.assertAlmostEqual(1,     f.iline[1][0][0], places = 4)
                self.assertAlmostEqual(3.004, f.iline[3][0][4], places = 4)
                self.assertAlmostEqual(3.014, f.iline[3][1][4], places = 4)
                self.assertAlmostEqual(7.023, f.iline[7][2][3], places = 4)

    def test_segyio_types(self):
        with segyio.open(self.filename, "r") as f:
            self.assertIsInstance(f.sorting, int)
            self.assertIsInstance(f.ext_headers, int)
            self.assertIsInstance(f.tracecount, int)
            self.assertIsInstance(f.samples, np.ndarray)

            self.assertIsInstance(f.depth_slice, Line)
            self.assertIsInstance(f.depth_slice[1], np.ndarray)
            self.assertIsInstance(f.depth_slice[1:23], GeneratorType)

            self.assertIsInstance(f.ilines, np.ndarray)
            self.assertIsInstance(f.iline, Line)
            self.assertIsInstance(f.iline[1], np.ndarray)
            self.assertIsInstance(f.iline[1:3], GeneratorType)
            self.assertIsInstance(f.iline[1][0], np.ndarray)
            self.assertIsInstance(f.iline[1][0:2], np.ndarray)
            self.assertIsInstance(float(f.iline[1][0][0]), float)
            self.assertIsInstance(f.iline[1][0][0:3], np.ndarray)

            self.assertIsInstance(f.xlines, np.ndarray)
            self.assertIsInstance(f.xline, Line)
            self.assertIsInstance(f.xline[21], np.ndarray)
            self.assertIsInstance(f.xline[21:23], GeneratorType)
            self.assertIsInstance(f.xline[21][0], np.ndarray)
            self.assertIsInstance(f.xline[21][0:2], np.ndarray)
            self.assertIsInstance(float(f.xline[21][0][0]), float)
            self.assertIsInstance(f.xline[21][0][0:3], np.ndarray)

            self.assertIsInstance(f.header, Header)
            self.assertIsInstance(f.header.iline, Line)
            self.assertIsInstance(f.header.iline[1], GeneratorType)
            self.assertIsInstance(next(f.header.iline[1]), Field)
            self.assertIsInstance(f.header.xline, Line)
            self.assertIsInstance(f.header.xline[21], GeneratorType)
            self.assertIsInstance(next(f.header.xline[21]), Field)

            self.assertIsInstance(f.trace, Trace)
            self.assertIsInstance(f.trace[0], np.ndarray)

            self.assertIsInstance(f.bin, Field)
            self.assertIsInstance(f.text, object)  # inner TextHeader instance

    def test_depth_slice_reading(self):
        with segyio.open(self.filename, "r") as f:
            self.assertEqual(len(f.depth_slice), len(f.samples))

            for depth_sample in range(len(f.samples)):
                depth_slice = f.depth_slice[depth_sample]
                self.assertIsInstance(depth_slice, np.ndarray)
                self.assertEqual(depth_slice.shape, (5, 5))

                for x, y in itertools.product(f.ilines, f.xlines):
                    i, j = x - f.ilines[0], y - f.xlines[0]
                    self.assertAlmostEqual(depth_slice[i][j], f.iline[x][j][depth_sample], places=6)

            for index, depth_slice in enumerate(f.depth_slice):
                self.assertIsInstance(depth_slice, np.ndarray)
                self.assertEqual(depth_slice.shape, (5, 5))

                for x, y in itertools.product(f.ilines, f.xlines):
                    i, j = x - f.ilines[0], y - f.xlines[0]
                    self.assertAlmostEqual(depth_slice[i][j], f.iline[x][j][index], places=6)

        with self.assertRaises(KeyError):
            slice = f.depth_slice[len(f.samples)]

    def test_depth_slice_writing(self):
        with TestContext("depth_slice_writing") as context:
            context.copy_file(self.filename)
            fname = ("small.sgy")

            buf = np.empty(shape=(5, 5), dtype=np.single)

            def value(x, y):
                return x + (1.0 // 5) * y

            for x, y in itertools.product(range(5), range(5)):
                buf[x][y] = value(x, y)

            with segyio.open(fname, "r+") as f:
                f.depth_slice[7] = buf * 3.14 # assign to depth 7
                self.assertTrue(np.allclose(f.depth_slice[7], buf * 3.14))

                f.depth_slice = [buf * i for i in range(len(f.depth_slice))]  # assign to all depths
                for index, depth_slice in enumerate(f.depth_slice):
                    self.assertTrue(np.allclose(depth_slice, buf * index))

if __name__ == '__main__':
    unittest.main()



import unittest
import subprocess
import os

class TestSegyH(unittest.TestCase):

    actual = [
        b"C 2 AN INCREASE IN AMPLITUDE EQUALS AN INCREASE IN ACOUSTIC IMPEDANCE",
        b"C 3 Written by libsegyio (python)",
        b"C 4",
        b"C 5",
        b"C 6",
        b"C 7",
        b"C 8",
        b"C 9",
        b"C10",
        b"C11 TRACE HEADER POSITION:",
        b"C12   INLINE BYTES 189-193    | OFFSET BYTES 037-041",
        b"C13   CROSSLINE BYTES 193-197 |",
        b"C14",
        b"C15 END EBCDIC HEADER",
    ]
    work_dir = '../'
    cmd_base = './applications/segyio-cath'
    FNULL = open(os.devnull, 'w')

    def test_segy_in(self):
        cmd = self.cmd_base + ' python/test-data/small.sgy'
        result = subprocess.check_output(cmd, shell=True,
                 cwd=self.work_dir).split(b'\n')
        for i in range(len(self.actual)):
            self.assertEqual(self.actual[i].strip(), b''+result[i + 1].strip())

    def test_none_segy_in(self):
        cmd = self.cmd_base + ' ../CMakeLists.txt'
        result = subprocess.call(cmd, shell=True,
                 stderr=self.FNULL, cwd=self.work_dir)
        self.assertEqual(result, 0)

if __name__ == '__main__':
    unittest.main()
